/* Generated by Cython 0.28.3 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_3"
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__Orange__data___io
#define __PYX_HAVE_API__Orange__data___io
/* Early includes */
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "Orange/data/_io.pyx",
};

/*--- Type declarations ---*/

/* "Orange/data/_io.pyx":7
 * 
 * 
 * cdef enum State:             # <<<<<<<<<<<<<<
 *     BEGIN_LINE, READ_START_ATOM, READ, QUOTED, END_QUOTED, END_ATOM,
 *     COMMENT, CARRIAGE_RETURNED, ESCAPE, EQUALS,
 */
enum __pyx_t_6Orange_4data_3_io_State {
  __pyx_e_6Orange_4data_3_io_BEGIN_LINE,
  __pyx_e_6Orange_4data_3_io_READ_START_ATOM,
  __pyx_e_6Orange_4data_3_io_READ,
  __pyx_e_6Orange_4data_3_io_QUOTED,
  __pyx_e_6Orange_4data_3_io_END_QUOTED,
  __pyx_e_6Orange_4data_3_io_END_ATOM,
  __pyx_e_6Orange_4data_3_io_COMMENT,
  __pyx_e_6Orange_4data_3_io_CARRIAGE_RETURNED,
  __pyx_e_6Orange_4data_3_io_ESCAPE,
  __pyx_e_6Orange_4data_3_io_EQUALS,
  __pyx_e_6Orange_4data_3_io_SET_VALUE,
  __pyx_e_6Orange_4data_3_io_WAIT_VALUE,
  __pyx_e_6Orange_4data_3_io_READ_VALUE,
  __pyx_e_6Orange_4data_3_io_READ_DECS,
  __pyx_e_6Orange_4data_3_io_TO_NEXT
};

/* "Orange/data/_io.pyx":12
 *     SET_VALUE, WAIT_VALUE, READ_VALUE, READ_DECS, TO_NEXT
 * 
 * cdef enum ColKinds:             # <<<<<<<<<<<<<<
 *     ATTRIBUTE, CLASS, META
 * 
 */
enum __pyx_t_6Orange_4data_3_io_ColKinds {
  __pyx_e_6Orange_4data_3_io_ATTRIBUTE,
  __pyx_e_6Orange_4data_3_io_CLASS,
  __pyx_e_6Orange_4data_3_io_META
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'Orange.data._io' */
#define __Pyx_MODULE_NAME "Orange.data._io"
extern int __pyx_module_is_main_Orange__data___io;
int __pyx_module_is_main_Orange__data___io = 0;

/* Implementation of 'Orange.data._io' */
static PyObject *__pyx_builtin_IOError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_[] = "*";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_ci[] = "ci";
static const char __pyx_k_ll[] = "ll";
static const char __pyx_k_sp[] = "sp";
static const char __pyx_k_col[] = "col";
static const char __pyx_k_mat[] = "mat";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_atom[] = "atom";
static const char __pyx_k_decs[] = "decs";
static const char __pyx_k_endc[] = "endc";
static const char __pyx_k_line[] = "line";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_atome[] = "atome";
static const char __pyx_k_atomp[] = "atomp";
static const char __pyx_k_f_eof[] = "f_eof";
static const char __pyx_k_fname[] = "fname";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_tocsr[] = "tocsr";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_M_cols[] = "M_cols";
static const char __pyx_k_M_data[] = "M_data";
static const char __pyx_k_M_rows[] = "M_rows";
static const char __pyx_k_X_cols[] = "X_cols";
static const char __pyx_k_X_data[] = "X_data";
static const char __pyx_k_X_rows[] = "X_rows";
static const char __pyx_k_Y_cols[] = "Y_cols";
static const char __pyx_k_Y_data[] = "Y_data";
static const char __pyx_k_Y_rows[] = "Y_rows";
static const char __pyx_k_b_atom[] = "b_atom";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_t_cols[] = "t_cols";
static const char __pyx_k_t_data[] = "t_data";
static const char __pyx_k_t_rows[] = "t_rows";
static const char __pyx_k_IOError[] = "IOError";
static const char __pyx_k_in_line[] = "in_line";
static const char __pyx_k_t_names[] = "t_names";
static const char __pyx_k_col_kind[] = "col_kind";
static const char __pyx_k_cur_line[] = "cur_line";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_attr_index[] = "attr_index";
static const char __pyx_k_coo_matrix[] = "coo_matrix";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_not_in_atom[] = "not_in_atom";
static const char __pyx_k_attr_indices[] = "attr_indices";
static const char __pyx_k_meta_indices[] = "meta_indices";
static const char __pyx_k_scipy_sparse[] = "scipy.sparse";
static const char __pyx_k_sort_indices[] = "sort_indices";
static const char __pyx_k_class_indices[] = "class_indices";
static const char __pyx_k_invalid_value[] = "{}:{}:{}: invalid value";
static const char __pyx_k_Orange_data__io[] = "Orange.data._io";
static const char __pyx_k_empty_value_name[] = "{}:{}:{}: empty value name";
static const char __pyx_k_sparse_read_float[] = "sparse_read_float";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_missing_value_name[] = "{}:{}:{}: missing value name";
static const char __pyx_k_Orange_data__io_pyx[] = "Orange/data/_io.pyx";
static const char __pyx_k_value_name_too_long[] = "{}:{}:{}: value name too long";
static const char __pyx_k_File_cannot_be_opened[] = "File '{}' cannot be opened";
static const char __pyx_k_duplicated_semi_colons[] = "{}:{}:{} duplicated semi-colons";
static const char __pyx_k_unrecognized_escape_sequence[] = "{}:{}:{}: unrecognized escape sequence";
static const char __pyx_k_end_of_file_in_escape_sequence[] = "{}:{}:{}: end of file in escape sequence";
static const char __pyx_k_end_of_line_in_escape_sequence[] = "{}:{}:{}: end of line in escape sequence";
static const char __pyx_k_classes_should_follow_attribute[] = "{}:{}:{}: classes should follow attributes";
static const char __pyx_k_end_of_file_within_a_quoted_val[] = "{}:{}:{}: end of file within a quoted value";
static const char __pyx_k_end_of_line_within_a_quoted_val[] = "{}:{}:{}: end of line within a quoted value";
static const char __pyx_k_quoted_value_should_be_followed[] = "{}:{}:{}: quoted value should be followed by value separator or end of line";
static PyObject *__pyx_n_s_;
static PyObject *__pyx_kp_u_File_cannot_be_opened;
static PyObject *__pyx_n_s_IOError;
static PyObject *__pyx_n_s_M_cols;
static PyObject *__pyx_n_s_M_data;
static PyObject *__pyx_n_s_M_rows;
static PyObject *__pyx_n_s_Orange_data__io;
static PyObject *__pyx_kp_s_Orange_data__io_pyx;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_X_cols;
static PyObject *__pyx_n_s_X_data;
static PyObject *__pyx_n_s_X_rows;
static PyObject *__pyx_n_s_Y_cols;
static PyObject *__pyx_n_s_Y_data;
static PyObject *__pyx_n_s_Y_rows;
static PyObject *__pyx_n_s_atom;
static PyObject *__pyx_n_s_atome;
static PyObject *__pyx_n_s_atomp;
static PyObject *__pyx_n_s_attr_index;
static PyObject *__pyx_n_s_attr_indices;
static PyObject *__pyx_n_s_b_atom;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_ci;
static PyObject *__pyx_n_s_class_indices;
static PyObject *__pyx_kp_u_classes_should_follow_attribute;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_col;
static PyObject *__pyx_n_s_col_kind;
static PyObject *__pyx_n_s_coo_matrix;
static PyObject *__pyx_n_s_cur_line;
static PyObject *__pyx_n_s_decs;
static PyObject *__pyx_kp_u_duplicated_semi_colons;
static PyObject *__pyx_kp_u_empty_value_name;
static PyObject *__pyx_kp_u_end_of_file_in_escape_sequence;
static PyObject *__pyx_kp_u_end_of_file_within_a_quoted_val;
static PyObject *__pyx_kp_u_end_of_line_in_escape_sequence;
static PyObject *__pyx_kp_u_end_of_line_within_a_quoted_val;
static PyObject *__pyx_n_s_endc;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_f_eof;
static PyObject *__pyx_n_s_fname;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_in_line;
static PyObject *__pyx_kp_u_invalid_value;
static PyObject *__pyx_n_s_line;
static PyObject *__pyx_n_s_ll;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mat;
static PyObject *__pyx_n_s_meta_indices;
static PyObject *__pyx_kp_u_missing_value_name;
static PyObject *__pyx_n_s_not_in_atom;
static PyObject *__pyx_kp_u_quoted_value_should_be_followed;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_scipy_sparse;
static PyObject *__pyx_n_s_setdefault;
static PyObject *__pyx_n_s_sort_indices;
static PyObject *__pyx_n_s_sp;
static PyObject *__pyx_n_s_sparse_read_float;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_t_cols;
static PyObject *__pyx_n_s_t_data;
static PyObject *__pyx_n_s_t_names;
static PyObject *__pyx_n_s_t_rows;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tocsr;
static PyObject *__pyx_kp_u_unrecognized_escape_sequence;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_kp_u_value_name_too_long;
static PyObject *__pyx_pf_6Orange_4data_3_io_sparse_read_float(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fname); /* proto */
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_setdefault = {0, &__pyx_n_s_setdefault, 0, 0, 0};
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_codeobj__3;
/* Late includes */

/* "Orange/data/_io.pyx":16
 * 
 * 
 * def sparse_read_float(fname):             # <<<<<<<<<<<<<<
 *     cdef:
 *         State state
 */

/* Python wrapper */
static PyObject *__pyx_pw_6Orange_4data_3_io_1sparse_read_float(PyObject *__pyx_self, PyObject *__pyx_v_fname); /*proto*/
static PyMethodDef __pyx_mdef_6Orange_4data_3_io_1sparse_read_float = {"sparse_read_float", (PyCFunction)__pyx_pw_6Orange_4data_3_io_1sparse_read_float, METH_O, 0};
static PyObject *__pyx_pw_6Orange_4data_3_io_1sparse_read_float(PyObject *__pyx_self, PyObject *__pyx_v_fname) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sparse_read_float (wrapper)", 0);
  __pyx_r = __pyx_pf_6Orange_4data_3_io_sparse_read_float(__pyx_self, ((PyObject *)__pyx_v_fname));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6Orange_4data_3_io_sparse_read_float(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fname) {
  enum __pyx_t_6Orange_4data_3_io_State __pyx_v_state;
  char __pyx_v_c;
  int __pyx_v_ci;
  char *__pyx_v_not_in_atom;
  int __pyx_v_col;
  int __pyx_v_line;
  int __pyx_v_cur_line;
  int __pyx_v_in_line;
  char __pyx_v_atom[0x2800];
  char *__pyx_v_atomp;
  char *__pyx_v_atome;
  char *__pyx_v_endc;
  char __pyx_v_f_eof;
  int __pyx_v_attr_index;
  float __pyx_v_value;
  float __pyx_v_decs;
  char __pyx_v_col_kind;
  PyObject *__pyx_v_X_data = 0;
  PyObject *__pyx_v_X_rows = 0;
  PyObject *__pyx_v_X_cols = 0;
  PyObject *__pyx_v_Y_data = 0;
  PyObject *__pyx_v_Y_rows = 0;
  PyObject *__pyx_v_Y_cols = 0;
  PyObject *__pyx_v_M_data = 0;
  PyObject *__pyx_v_M_rows = 0;
  PyObject *__pyx_v_M_cols = 0;
  PyObject *__pyx_v_attr_indices = 0;
  PyObject *__pyx_v_class_indices = 0;
  PyObject *__pyx_v_meta_indices = 0;
  FILE *__pyx_v_f;
  PyObject *__pyx_v_t_names = NULL;
  char *__pyx_v_b_atom;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_t_data = NULL;
  PyObject *__pyx_v_t_rows = NULL;
  PyObject *__pyx_v_t_cols = NULL;
  PyObject *__pyx_v_ll = NULL;
  PyObject *__pyx_v_mat = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  Py_ssize_t __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  __Pyx_RefNannySetupContext("sparse_read_float", 0);

  /* "Orange/data/_io.pyx":21
 *         char c
 *         int ci
 *         char *not_in_atom = "#,|;\n\r\x00"             # <<<<<<<<<<<<<<
 *         int col, line, cur_line
 *         int in_line
 */
  __pyx_v_not_in_atom = ((char *)"#,|;\n\r\000");

  /* "Orange/data/_io.pyx":26
 *         char atom[10240]
 *         char *atomp
 *         char *atome = atom + 10240             # <<<<<<<<<<<<<<
 *         char *endc
 *         char f_eof = 0
 */
  __pyx_v_atome = (__pyx_v_atom + 0x2800);

  /* "Orange/data/_io.pyx":28
 *         char *atome = atom + 10240
 *         char *endc
 *         char f_eof = 0             # <<<<<<<<<<<<<<
 *         int attr_index
 *         float value, decs
 */
  __pyx_v_f_eof = 0;

  /* "Orange/data/_io.pyx":33
 *         char col_kind
 * 
 *         list X_data = []             # <<<<<<<<<<<<<<
 *         list X_rows = []
 *         list X_cols = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_X_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":34
 * 
 *         list X_data = []
 *         list X_rows = []             # <<<<<<<<<<<<<<
 *         list X_cols = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_X_rows = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":35
 *         list X_data = []
 *         list X_rows = []
 *         list X_cols = []             # <<<<<<<<<<<<<<
 * 
 *         list Y_data = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_X_cols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":37
 *         list X_cols = []
 * 
 *         list Y_data = []             # <<<<<<<<<<<<<<
 *         list Y_rows = []
 *         list Y_cols = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_Y_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":38
 * 
 *         list Y_data = []
 *         list Y_rows = []             # <<<<<<<<<<<<<<
 *         list Y_cols = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_Y_rows = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":39
 *         list Y_data = []
 *         list Y_rows = []
 *         list Y_cols = []             # <<<<<<<<<<<<<<
 * 
 *         list M_data = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_Y_cols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":41
 *         list Y_cols = []
 * 
 *         list M_data = []             # <<<<<<<<<<<<<<
 *         list M_rows = []
 *         list M_cols = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_M_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":42
 * 
 *         list M_data = []
 *         list M_rows = []             # <<<<<<<<<<<<<<
 *         list M_cols = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_M_rows = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":43
 *         list M_data = []
 *         list M_rows = []
 *         list M_cols = []             # <<<<<<<<<<<<<<
 * 
 *         dict attr_indices = {}
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_M_cols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":45
 *         list M_cols = []
 * 
 *         dict attr_indices = {}             # <<<<<<<<<<<<<<
 *         dict class_indices = {}
 *         dict meta_indices = {}
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_attr_indices = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":46
 * 
 *         dict attr_indices = {}
 *         dict class_indices = {}             # <<<<<<<<<<<<<<
 *         dict meta_indices = {}
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_class_indices = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":47
 *         dict attr_indices = {}
 *         dict class_indices = {}
 *         dict meta_indices = {}             # <<<<<<<<<<<<<<
 * 
 *     cdef FILE *f = fopen(fname, "rb")
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_meta_indices = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Orange/data/_io.pyx":49
 *         dict meta_indices = {}
 * 
 *     cdef FILE *f = fopen(fname, "rb")             # <<<<<<<<<<<<<<
 *     if f == NULL:
 *         raise IOError("File '{}' cannot be opened".format(fname))
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_fname); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_v_f = fopen(__pyx_t_2, ((char const *)"rb"));

  /* "Orange/data/_io.pyx":50
 * 
 *     cdef FILE *f = fopen(fname, "rb")
 *     if f == NULL:             # <<<<<<<<<<<<<<
 *         raise IOError("File '{}' cannot be opened".format(fname))
 * 
 */
  __pyx_t_3 = ((__pyx_v_f == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "Orange/data/_io.pyx":51
 *     cdef FILE *f = fopen(fname, "rb")
 *     if f == NULL:
 *         raise IOError("File '{}' cannot be opened".format(fname))             # <<<<<<<<<<<<<<
 * 
 *     line = 0
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_File_cannot_be_opened, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_fname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_fname};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_fname};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 51, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_fname);
        __Pyx_GIVEREF(__pyx_v_fname);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_fname);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 51, __pyx_L1_error)

    /* "Orange/data/_io.pyx":50
 * 
 *     cdef FILE *f = fopen(fname, "rb")
 *     if f == NULL:             # <<<<<<<<<<<<<<
 *         raise IOError("File '{}' cannot be opened".format(fname))
 * 
 */
  }

  /* "Orange/data/_io.pyx":53
 *         raise IOError("File '{}' cannot be opened".format(fname))
 * 
 *     line = 0             # <<<<<<<<<<<<<<
 *     cur_line = 0
 *     in_line = 0
 */
  __pyx_v_line = 0;

  /* "Orange/data/_io.pyx":54
 * 
 *     line = 0
 *     cur_line = 0             # <<<<<<<<<<<<<<
 *     in_line = 0
 * 
 */
  __pyx_v_cur_line = 0;

  /* "Orange/data/_io.pyx":55
 *     line = 0
 *     cur_line = 0
 *     in_line = 0             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_v_in_line = 0;

  /* "Orange/data/_io.pyx":57
 *     in_line = 0
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         state = BEGIN_LINE
 *         while not (f_eof and state == BEGIN_LINE):
 */
  /*try:*/ {

    /* "Orange/data/_io.pyx":58
 * 
 *     try:
 *         state = BEGIN_LINE             # <<<<<<<<<<<<<<
 *         while not (f_eof and state == BEGIN_LINE):
 *             if state == BEGIN_LINE:
 */
    __pyx_v_state = __pyx_e_6Orange_4data_3_io_BEGIN_LINE;

    /* "Orange/data/_io.pyx":59
 *     try:
 *         state = BEGIN_LINE
 *         while not (f_eof and state == BEGIN_LINE):             # <<<<<<<<<<<<<<
 *             if state == BEGIN_LINE:
 *                 col_kind = ATTRIBUTE
 */
    while (1) {
      __pyx_t_7 = (__pyx_v_f_eof != 0);
      if (__pyx_t_7) {
      } else {
        __pyx_t_3 = __pyx_t_7;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_BEGIN_LINE) != 0);
      __pyx_t_3 = __pyx_t_7;
      __pyx_L9_bool_binop_done:;
      __pyx_t_7 = ((!__pyx_t_3) != 0);
      if (!__pyx_t_7) break;

      /* "Orange/data/_io.pyx":60
 *         state = BEGIN_LINE
 *         while not (f_eof and state == BEGIN_LINE):
 *             if state == BEGIN_LINE:             # <<<<<<<<<<<<<<
 *                 col_kind = ATTRIBUTE
 *                 if in_line or line == 0:
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_BEGIN_LINE) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":61
 *         while not (f_eof and state == BEGIN_LINE):
 *             if state == BEGIN_LINE:
 *                 col_kind = ATTRIBUTE             # <<<<<<<<<<<<<<
 *                 if in_line or line == 0:
 *                     line += 1
 */
        __pyx_v_col_kind = __pyx_e_6Orange_4data_3_io_ATTRIBUTE;

        /* "Orange/data/_io.pyx":62
 *             if state == BEGIN_LINE:
 *                 col_kind = ATTRIBUTE
 *                 if in_line or line == 0:             # <<<<<<<<<<<<<<
 *                     line += 1
 *                     t_names = attr_indices
 */
        __pyx_t_3 = (__pyx_v_in_line != 0);
        if (!__pyx_t_3) {
        } else {
          __pyx_t_7 = __pyx_t_3;
          goto __pyx_L13_bool_binop_done;
        }
        __pyx_t_3 = ((__pyx_v_line == 0) != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L13_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":63
 *                 col_kind = ATTRIBUTE
 *                 if in_line or line == 0:
 *                     line += 1             # <<<<<<<<<<<<<<
 *                     t_names = attr_indices
 *                 cur_line += 1
 */
          __pyx_v_line = (__pyx_v_line + 1);

          /* "Orange/data/_io.pyx":64
 *                 if in_line or line == 0:
 *                     line += 1
 *                     t_names = attr_indices             # <<<<<<<<<<<<<<
 *                 cur_line += 1
 *                 col = in_line = 0
 */
          __Pyx_INCREF(__pyx_v_attr_indices);
          __Pyx_XDECREF_SET(__pyx_v_t_names, __pyx_v_attr_indices);

          /* "Orange/data/_io.pyx":62
 *             if state == BEGIN_LINE:
 *                 col_kind = ATTRIBUTE
 *                 if in_line or line == 0:             # <<<<<<<<<<<<<<
 *                     line += 1
 *                     t_names = attr_indices
 */
        }

        /* "Orange/data/_io.pyx":65
 *                     line += 1
 *                     t_names = attr_indices
 *                 cur_line += 1             # <<<<<<<<<<<<<<
 *                 col = in_line = 0
 *                 state = READ_START_ATOM
 */
        __pyx_v_cur_line = (__pyx_v_cur_line + 1);

        /* "Orange/data/_io.pyx":66
 *                     t_names = attr_indices
 *                 cur_line += 1
 *                 col = in_line = 0             # <<<<<<<<<<<<<<
 *                 state = READ_START_ATOM
 * 
 */
        __pyx_v_col = 0;
        __pyx_v_in_line = 0;

        /* "Orange/data/_io.pyx":67
 *                 cur_line += 1
 *                 col = in_line = 0
 *                 state = READ_START_ATOM             # <<<<<<<<<<<<<<
 * 
 *             if state != SET_VALUE:
 */
        __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_START_ATOM;

        /* "Orange/data/_io.pyx":60
 *         state = BEGIN_LINE
 *         while not (f_eof and state == BEGIN_LINE):
 *             if state == BEGIN_LINE:             # <<<<<<<<<<<<<<
 *                 col_kind = ATTRIBUTE
 *                 if in_line or line == 0:
 */
      }

      /* "Orange/data/_io.pyx":69
 *                 state = READ_START_ATOM
 * 
 *             if state != SET_VALUE:             # <<<<<<<<<<<<<<
 *                 ci = fgetc(f)
 *                 if ci == EOF:
 */
      __pyx_t_7 = ((__pyx_v_state != __pyx_e_6Orange_4data_3_io_SET_VALUE) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":70
 * 
 *             if state != SET_VALUE:
 *                 ci = fgetc(f)             # <<<<<<<<<<<<<<
 *                 if ci == EOF:
 *                     f_eof = 1
 */
        __pyx_v_ci = fgetc(__pyx_v_f);

        /* "Orange/data/_io.pyx":71
 *             if state != SET_VALUE:
 *                 ci = fgetc(f)
 *                 if ci == EOF:             # <<<<<<<<<<<<<<
 *                     f_eof = 1
 *                     c = b"\x00"
 */
        __pyx_t_7 = ((__pyx_v_ci == EOF) != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":72
 *                 ci = fgetc(f)
 *                 if ci == EOF:
 *                     f_eof = 1             # <<<<<<<<<<<<<<
 *                     c = b"\x00"
 *                 else:
 */
          __pyx_v_f_eof = 1;

          /* "Orange/data/_io.pyx":73
 *                 if ci == EOF:
 *                     f_eof = 1
 *                     c = b"\x00"             # <<<<<<<<<<<<<<
 *                 else:
 *                     c = <char>ci
 */
          __pyx_v_c = '\x00';

          /* "Orange/data/_io.pyx":71
 *             if state != SET_VALUE:
 *                 ci = fgetc(f)
 *                 if ci == EOF:             # <<<<<<<<<<<<<<
 *                     f_eof = 1
 *                     c = b"\x00"
 */
          goto __pyx_L16;
        }

        /* "Orange/data/_io.pyx":75
 *                     c = b"\x00"
 *                 else:
 *                     c = <char>ci             # <<<<<<<<<<<<<<
 *                 col += 1
 * 
 */
        /*else*/ {
          __pyx_v_c = ((char)__pyx_v_ci);
        }
        __pyx_L16:;

        /* "Orange/data/_io.pyx":76
 *                 else:
 *                     c = <char>ci
 *                 col += 1             # <<<<<<<<<<<<<<
 * 
 *             if state == READ_START_ATOM:
 */
        __pyx_v_col = (__pyx_v_col + 1);

        /* "Orange/data/_io.pyx":69
 *                 state = READ_START_ATOM
 * 
 *             if state != SET_VALUE:             # <<<<<<<<<<<<<<
 *                 ci = fgetc(f)
 *                 if ci == EOF:
 */
      }

      /* "Orange/data/_io.pyx":78
 *                 col += 1
 * 
 *             if state == READ_START_ATOM:             # <<<<<<<<<<<<<<
 *                 atomp = atom
 *                 value = 1
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_READ_START_ATOM) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":79
 * 
 *             if state == READ_START_ATOM:
 *                 atomp = atom             # <<<<<<<<<<<<<<
 *                 value = 1
 *                 if c == b"," or c == b" " or c == b"\t":
 */
        __pyx_v_atomp = __pyx_v_atom;

        /* "Orange/data/_io.pyx":80
 *             if state == READ_START_ATOM:
 *                 atomp = atom
 *                 value = 1             # <<<<<<<<<<<<<<
 *                 if c == b"," or c == b" " or c == b"\t":
 *                     continue
 */
        __pyx_v_value = 1.0;

        /* "Orange/data/_io.pyx":81
 *                 atomp = atom
 *                 value = 1
 *                 if c == b"," or c == b" " or c == b"\t":             # <<<<<<<<<<<<<<
 *                     continue
 *                 elif c == b'"':
 */
        switch (__pyx_v_c) {
          case ',':
          case ' ':
          case '\t':

          /* "Orange/data/_io.pyx":82
 *                 value = 1
 *                 if c == b"," or c == b" " or c == b"\t":
 *                     continue             # <<<<<<<<<<<<<<
 *                 elif c == b'"':
 *                     state = QUOTED
 */
          goto __pyx_L7_continue;

          /* "Orange/data/_io.pyx":81
 *                 atomp = atom
 *                 value = 1
 *                 if c == b"," or c == b" " or c == b"\t":             # <<<<<<<<<<<<<<
 *                     continue
 *                 elif c == b'"':
 */
          break;

          /* "Orange/data/_io.pyx":83
 *                 if c == b"," or c == b" " or c == b"\t":
 *                     continue
 *                 elif c == b'"':             # <<<<<<<<<<<<<<
 *                     state = QUOTED
 *                     continue
 */
          case '"':

          /* "Orange/data/_io.pyx":84
 *                     continue
 *                 elif c == b'"':
 *                     state = QUOTED             # <<<<<<<<<<<<<<
 *                     continue
 *                 elif c == b"=":
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_QUOTED;

          /* "Orange/data/_io.pyx":85
 *                 elif c == b'"':
 *                     state = QUOTED
 *                     continue             # <<<<<<<<<<<<<<
 *                 elif c == b"=":
 *                     raise ValueError("{}:{}:{}: missing value name"
 */
          goto __pyx_L7_continue;

          /* "Orange/data/_io.pyx":83
 *                 if c == b"," or c == b" " or c == b"\t":
 *                     continue
 *                 elif c == b'"':             # <<<<<<<<<<<<<<
 *                     state = QUOTED
 *                     continue
 */
          break;

          /* "Orange/data/_io.pyx":86
 *                     state = QUOTED
 *                     continue
 *                 elif c == b"=":             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: missing value name"
 *                         .format(fname, cur_line, col))
 */
          case '=':

          /* "Orange/data/_io.pyx":88
 *                 elif c == b"=":
 *                     raise ValueError("{}:{}:{}: missing value name"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 # fall through
 *                 state = READ
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_missing_value_name, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 88, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 88, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_6 = 0;
            __pyx_t_5 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Orange/data/_io.pyx":87
 *                     continue
 *                 elif c == b"=":
 *                     raise ValueError("{}:{}:{}: missing value name"             # <<<<<<<<<<<<<<
 *                         .format(fname, cur_line, col))
 *                 # fall through
 */
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 87, __pyx_L5_error)

          /* "Orange/data/_io.pyx":86
 *                     state = QUOTED
 *                     continue
 *                 elif c == b"=":             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: missing value name"
 *                         .format(fname, cur_line, col))
 */
          break;
          default: break;
        }

        /* "Orange/data/_io.pyx":90
 *                         .format(fname, cur_line, col))
 *                 # fall through
 *                 state = READ             # <<<<<<<<<<<<<<
 * 
 *             if state == ESCAPE:
 */
        __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ;

        /* "Orange/data/_io.pyx":78
 *                 col += 1
 * 
 *             if state == READ_START_ATOM:             # <<<<<<<<<<<<<<
 *                 atomp = atom
 *                 value = 1
 */
      }

      /* "Orange/data/_io.pyx":92
 *                 state = READ
 * 
 *             if state == ESCAPE:             # <<<<<<<<<<<<<<
 *                 if c == b"t":    c = b"\t"
 *                 elif c == b"n":    c = b"\n"
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_ESCAPE) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":93
 * 
 *             if state == ESCAPE:
 *                 if c == b"t":    c = b"\t"             # <<<<<<<<<<<<<<
 *                 elif c == b"n":    c = b"\n"
 *                 elif c == b"r":    c = b"\r"
 */
        __pyx_t_7 = ((__pyx_v_c == 't') != 0);
        if (__pyx_t_7) {
          __pyx_v_c = '\t';
          goto __pyx_L19;
        }

        /* "Orange/data/_io.pyx":94
 *             if state == ESCAPE:
 *                 if c == b"t":    c = b"\t"
 *                 elif c == b"n":    c = b"\n"             # <<<<<<<<<<<<<<
 *                 elif c == b"r":    c = b"\r"
 *                 elif c == b'"' or c == b"'" or c == b"\\" or c == b" ":
 */
        __pyx_t_7 = ((__pyx_v_c == 'n') != 0);
        if (__pyx_t_7) {
          __pyx_v_c = '\n';
          goto __pyx_L19;
        }

        /* "Orange/data/_io.pyx":95
 *                 if c == b"t":    c = b"\t"
 *                 elif c == b"n":    c = b"\n"
 *                 elif c == b"r":    c = b"\r"             # <<<<<<<<<<<<<<
 *                 elif c == b'"' or c == b"'" or c == b"\\" or c == b" ":
 *                     pass
 */
        __pyx_t_7 = ((__pyx_v_c == 'r') != 0);
        if (__pyx_t_7) {
          __pyx_v_c = '\r';
          goto __pyx_L19;
        }

        /* "Orange/data/_io.pyx":96
 *                 elif c == b"n":    c = b"\n"
 *                 elif c == b"r":    c = b"\r"
 *                 elif c == b'"' or c == b"'" or c == b"\\" or c == b" ":             # <<<<<<<<<<<<<<
 *                     pass
 *                 elif c == b"\r" or c == b"\n":
 */
        switch (__pyx_v_c) {
          case '"':
          case '\'':
          case '\\':
          case ' ':
          __pyx_t_7 = 1;
          break;
          default:
          __pyx_t_7 = 0;
          break;
        }
        if (__pyx_t_7) {
          goto __pyx_L19;
        }

        /* "Orange/data/_io.pyx":98
 *                 elif c == b'"' or c == b"'" or c == b"\\" or c == b" ":
 *                     pass
 *                 elif c == b"\r" or c == b"\n":             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: end of line in escape sequence"
 *                         .format(fname, cur_line, col))
 */
        switch (__pyx_v_c) {
          case '\r':
          case '\n':
          __pyx_t_7 = 1;
          break;
          default:
          __pyx_t_7 = 0;
          break;
        }
        if (unlikely(__pyx_t_7)) {

          /* "Orange/data/_io.pyx":100
 *                 elif c == b"\r" or c == b"\n":
 *                     raise ValueError("{}:{}:{}: end of line in escape sequence"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 elif f_eof:
 *                     raise ValueError("{}:{}:{}: end of file in escape sequence"
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_end_of_line_in_escape_sequence, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 100, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 100, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_10, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_10, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_8 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 100, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_8);
            if (__pyx_t_6) {
              __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_10);
            PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_10 = 0;
            __pyx_t_5 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Orange/data/_io.pyx":99
 *                     pass
 *                 elif c == b"\r" or c == b"\n":
 *                     raise ValueError("{}:{}:{}: end of line in escape sequence"             # <<<<<<<<<<<<<<
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:
 */
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 99, __pyx_L5_error)

          /* "Orange/data/_io.pyx":98
 *                 elif c == b'"' or c == b"'" or c == b"\\" or c == b" ":
 *                     pass
 *                 elif c == b"\r" or c == b"\n":             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: end of line in escape sequence"
 *                         .format(fname, cur_line, col))
 */
        }

        /* "Orange/data/_io.pyx":101
 *                     raise ValueError("{}:{}:{}: end of line in escape sequence"
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: end of file in escape sequence"
 *                         .format(fname, cur_line, col))
 */
        __pyx_t_7 = (__pyx_v_f_eof != 0);
        if (unlikely(__pyx_t_7)) {

          /* "Orange/data/_io.pyx":103
 *                 elif f_eof:
 *                     raise ValueError("{}:{}:{}: end of file in escape sequence"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError("{}:{}:{}: unrecognized escape sequence"
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_end_of_file_in_escape_sequence, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 103, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 103, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 103, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 103, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_6 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 103, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_6);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_8 = 0;
            __pyx_t_5 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 103, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Orange/data/_io.pyx":102
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:
 *                     raise ValueError("{}:{}:{}: end of file in escape sequence"             # <<<<<<<<<<<<<<
 *                         .format(fname, cur_line, col))
 *                 else:
 */
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 102, __pyx_L5_error)

          /* "Orange/data/_io.pyx":101
 *                     raise ValueError("{}:{}:{}: end of line in escape sequence"
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: end of file in escape sequence"
 *                         .format(fname, cur_line, col))
 */
        }

        /* "Orange/data/_io.pyx":105
 *                         .format(fname, cur_line, col))
 *                 else:
 *                     raise ValueError("{}:{}:{}: unrecognized escape sequence"             # <<<<<<<<<<<<<<
 *                         .format(fname, cur_line, col))
 *                 # fall through
 */
        /*else*/ {

          /* "Orange/data/_io.pyx":106
 *                 else:
 *                     raise ValueError("{}:{}:{}: unrecognized escape sequence"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 # fall through
 *                 state = READ
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unrecognized_escape_sequence, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 106, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 106, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 106, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 106, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_6 = 0;
            __pyx_t_5 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Orange/data/_io.pyx":105
 *                         .format(fname, cur_line, col))
 *                 else:
 *                     raise ValueError("{}:{}:{}: unrecognized escape sequence"             # <<<<<<<<<<<<<<
 *                         .format(fname, cur_line, col))
 *                 # fall through
 */
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 105, __pyx_L5_error)
        }
        __pyx_L19:;

        /* "Orange/data/_io.pyx":108
 *                         .format(fname, cur_line, col))
 *                 # fall through
 *                 state = READ             # <<<<<<<<<<<<<<
 * 
 *             if state == READ:
 */
        __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ;

        /* "Orange/data/_io.pyx":92
 *                 state = READ
 * 
 *             if state == ESCAPE:             # <<<<<<<<<<<<<<
 *                 if c == b"t":    c = b"\t"
 *                 elif c == b"n":    c = b"\n"
 */
      }

      /* "Orange/data/_io.pyx":110
 *                 state = READ
 * 
 *             if state == READ:             # <<<<<<<<<<<<<<
 *                 if c == b"\\":
 *                     state = ESCAPE
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_READ) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":111
 * 
 *             if state == READ:
 *                 if c == b"\\":             # <<<<<<<<<<<<<<
 *                     state = ESCAPE
 *                     continue
 */
        __pyx_t_7 = ((__pyx_v_c == '\\') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":112
 *             if state == READ:
 *                 if c == b"\\":
 *                     state = ESCAPE             # <<<<<<<<<<<<<<
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_ESCAPE;

          /* "Orange/data/_io.pyx":113
 *                 if c == b"\\":
 *                     state = ESCAPE
 *                     continue             # <<<<<<<<<<<<<<
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":
 */
          goto __pyx_L7_continue;

          /* "Orange/data/_io.pyx":111
 * 
 *             if state == READ:
 *                 if c == b"\\":             # <<<<<<<<<<<<<<
 *                     state = ESCAPE
 *                     continue
 */
        }

        /* "Orange/data/_io.pyx":114
 *                     state = ESCAPE
 *                     continue
 *                 endc = strchr(not_in_atom, c)             # <<<<<<<<<<<<<<
 *                 if endc == NULL and c != b"=":
 *                     atomp[0] = c
 */
        __pyx_v_endc = strchr(__pyx_v_not_in_atom, __pyx_v_c);

        /* "Orange/data/_io.pyx":115
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":             # <<<<<<<<<<<<<<
 *                     atomp[0] = c
 *                     atomp += 1
 */
        __pyx_t_3 = ((__pyx_v_endc == NULL) != 0);
        if (__pyx_t_3) {
        } else {
          __pyx_t_7 = __pyx_t_3;
          goto __pyx_L23_bool_binop_done;
        }
        __pyx_t_3 = ((__pyx_v_c != '=') != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L23_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":116
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":
 *                     atomp[0] = c             # <<<<<<<<<<<<<<
 *                     atomp += 1
 *                     if atomp == atome:
 */
          (__pyx_v_atomp[0]) = __pyx_v_c;

          /* "Orange/data/_io.pyx":117
 *                 if endc == NULL and c != b"=":
 *                     atomp[0] = c
 *                     atomp += 1             # <<<<<<<<<<<<<<
 *                     if atomp == atome:
 *                         raise ValueError("{}:{}:{}: value name too long"
 */
          __pyx_v_atomp = (__pyx_v_atomp + 1);

          /* "Orange/data/_io.pyx":118
 *                     atomp[0] = c
 *                     atomp += 1
 *                     if atomp == atome:             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))
 */
          __pyx_t_7 = ((__pyx_v_atomp == __pyx_v_atome) != 0);
          if (unlikely(__pyx_t_7)) {

            /* "Orange/data/_io.pyx":120
 *                     if atomp == atome:
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_value_name_too_long, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 120, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 120, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_10, __pyx_t_5};
              __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_10, __pyx_t_5};
              __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 120, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (__pyx_t_6) {
                __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_10);
              PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_10);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_10 = 0;
              __pyx_t_5 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "Orange/data/_io.pyx":119
 *                     atomp += 1
 *                     if atomp == atome:
 *                         raise ValueError("{}:{}:{}: value name too long"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                     continue
 */
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 119, __pyx_L5_error)

            /* "Orange/data/_io.pyx":118
 *                     atomp[0] = c
 *                     atomp += 1
 *                     if atomp == atome:             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))
 */
          }

          /* "Orange/data/_io.pyx":121
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))
 *                     continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     # fall through to END_ATOM
 */
          goto __pyx_L7_continue;

          /* "Orange/data/_io.pyx":115
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":             # <<<<<<<<<<<<<<
 *                     atomp[0] = c
 *                     atomp += 1
 */
        }

        /* "Orange/data/_io.pyx":124
 *                 else:
 *                     # fall through to END_ATOM
 *                     state = END_ATOM             # <<<<<<<<<<<<<<
 * 
 *             if state == QUOTED:
 */
        /*else*/ {
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_END_ATOM;
        }

        /* "Orange/data/_io.pyx":110
 *                 state = READ
 * 
 *             if state == READ:             # <<<<<<<<<<<<<<
 *                 if c == b"\\":
 *                     state = ESCAPE
 */
      }

      /* "Orange/data/_io.pyx":126
 *                     state = END_ATOM
 * 
 *             if state == QUOTED:             # <<<<<<<<<<<<<<
 *                 if c == b"\r" or c == b"\n":
 *                     raise ValueError(
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_QUOTED) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":127
 * 
 *             if state == QUOTED:
 *                 if c == b"\r" or c == b"\n":             # <<<<<<<<<<<<<<
 *                     raise ValueError(
 *                         "{}:{}:{}: end of line within a quoted value"
 */
        switch (__pyx_v_c) {
          case '\r':
          case '\n':
          __pyx_t_7 = 1;
          break;
          default:
          __pyx_t_7 = 0;
          break;
        }
        if (unlikely(__pyx_t_7)) {

          /* "Orange/data/_io.pyx":130
 *                     raise ValueError(
 *                         "{}:{}:{}: end of line within a quoted value"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 elif f_eof:
 *                     raise ValueError(
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_end_of_line_within_a_quoted_val, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 130, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 130, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_6 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 130, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_6);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_8 = 0;
            __pyx_t_5 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Orange/data/_io.pyx":128
 *             if state == QUOTED:
 *                 if c == b"\r" or c == b"\n":
 *                     raise ValueError(             # <<<<<<<<<<<<<<
 *                         "{}:{}:{}: end of line within a quoted value"
 *                         .format(fname, cur_line, col))
 */
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 128, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 128, __pyx_L5_error)

          /* "Orange/data/_io.pyx":127
 * 
 *             if state == QUOTED:
 *                 if c == b"\r" or c == b"\n":             # <<<<<<<<<<<<<<
 *                     raise ValueError(
 *                         "{}:{}:{}: end of line within a quoted value"
 */
        }

        /* "Orange/data/_io.pyx":131
 *                         "{}:{}:{}: end of line within a quoted value"
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:             # <<<<<<<<<<<<<<
 *                     raise ValueError(
 *                         "{}:{}:{}: end of file within a quoted value"
 */
        __pyx_t_7 = (__pyx_v_f_eof != 0);
        if (unlikely(__pyx_t_7)) {

          /* "Orange/data/_io.pyx":134
 *                     raise ValueError(
 *                         "{}:{}:{}: end of file within a quoted value"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 elif c != b'"':
 *                     atomp[0] = c
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_end_of_file_within_a_quoted_val, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 134, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 134, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 134, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 134, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 134, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_6 = 0;
            __pyx_t_5 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 134, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Orange/data/_io.pyx":132
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:
 *                     raise ValueError(             # <<<<<<<<<<<<<<
 *                         "{}:{}:{}: end of file within a quoted value"
 *                         .format(fname, cur_line, col))
 */
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 132, __pyx_L5_error)

          /* "Orange/data/_io.pyx":131
 *                         "{}:{}:{}: end of line within a quoted value"
 *                         .format(fname, cur_line, col))
 *                 elif f_eof:             # <<<<<<<<<<<<<<
 *                     raise ValueError(
 *                         "{}:{}:{}: end of file within a quoted value"
 */
        }

        /* "Orange/data/_io.pyx":135
 *                         "{}:{}:{}: end of file within a quoted value"
 *                         .format(fname, cur_line, col))
 *                 elif c != b'"':             # <<<<<<<<<<<<<<
 *                     atomp[0] = c
 *                     atomp += 1
 */
        __pyx_t_7 = ((__pyx_v_c != '"') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":136
 *                         .format(fname, cur_line, col))
 *                 elif c != b'"':
 *                     atomp[0] = c             # <<<<<<<<<<<<<<
 *                     atomp += 1
 *                     if atomp == atome:
 */
          (__pyx_v_atomp[0]) = __pyx_v_c;

          /* "Orange/data/_io.pyx":137
 *                 elif c != b'"':
 *                     atomp[0] = c
 *                     atomp += 1             # <<<<<<<<<<<<<<
 *                     if atomp == atome:
 *                         raise ValueError("{}:{}:{}: value name too long"
 */
          __pyx_v_atomp = (__pyx_v_atomp + 1);

          /* "Orange/data/_io.pyx":138
 *                     atomp[0] = c
 *                     atomp += 1
 *                     if atomp == atome:             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))
 */
          __pyx_t_7 = ((__pyx_v_atomp == __pyx_v_atome) != 0);
          if (unlikely(__pyx_t_7)) {

            /* "Orange/data/_io.pyx":140
 *                     if atomp == atome:
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 else:
 *                     state = END_QUOTED
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_value_name_too_long, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 140, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 140, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_10, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_10, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 140, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (__pyx_t_6) {
                __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_10);
              PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_10);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_10 = 0;
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "Orange/data/_io.pyx":139
 *                     atomp += 1
 *                     if atomp == atome:
 *                         raise ValueError("{}:{}:{}: value name too long"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                 else:
 */
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(0, 139, __pyx_L5_error)

            /* "Orange/data/_io.pyx":138
 *                     atomp[0] = c
 *                     atomp += 1
 *                     if atomp == atome:             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{}: value name too long"
 *                             .format(fname, cur_line, col))
 */
          }

          /* "Orange/data/_io.pyx":135
 *                         "{}:{}:{}: end of file within a quoted value"
 *                         .format(fname, cur_line, col))
 *                 elif c != b'"':             # <<<<<<<<<<<<<<
 *                     atomp[0] = c
 *                     atomp += 1
 */
          goto __pyx_L27;
        }

        /* "Orange/data/_io.pyx":142
 *                             .format(fname, cur_line, col))
 *                 else:
 *                     state = END_QUOTED             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
        /*else*/ {
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_END_QUOTED;

          /* "Orange/data/_io.pyx":143
 *                 else:
 *                     state = END_QUOTED
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *             if state == END_QUOTED:
 */
          goto __pyx_L7_continue;
        }
        __pyx_L27:;

        /* "Orange/data/_io.pyx":126
 *                     state = END_ATOM
 * 
 *             if state == QUOTED:             # <<<<<<<<<<<<<<
 *                 if c == b"\r" or c == b"\n":
 *                     raise ValueError(
 */
      }

      /* "Orange/data/_io.pyx":145
 *                     continue
 * 
 *             if state == END_QUOTED:             # <<<<<<<<<<<<<<
 *                 if c == b" " or c == b"\t":
 *                     continue
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_END_QUOTED) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":146
 * 
 *             if state == END_QUOTED:
 *                 if c == b" " or c == b"\t":             # <<<<<<<<<<<<<<
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 */
        switch (__pyx_v_c) {
          case ' ':
          case '\t':

          /* "Orange/data/_io.pyx":147
 *             if state == END_QUOTED:
 *                 if c == b" " or c == b"\t":
 *                     continue             # <<<<<<<<<<<<<<
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":
 */
          goto __pyx_L7_continue;

          /* "Orange/data/_io.pyx":146
 * 
 *             if state == END_QUOTED:
 *                 if c == b" " or c == b"\t":             # <<<<<<<<<<<<<<
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 */
          break;
          default: break;
        }

        /* "Orange/data/_io.pyx":148
 *                 if c == b" " or c == b"\t":
 *                     continue
 *                 endc = strchr(not_in_atom, c)             # <<<<<<<<<<<<<<
 *                 if endc == NULL and c != b"=":
 *                     raise ValueError("{}:{}:{}: quoted value should be "
 */
        __pyx_v_endc = strchr(__pyx_v_not_in_atom, __pyx_v_c);

        /* "Orange/data/_io.pyx":149
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: quoted value should be "
 *                         "followed by value separator or end of line"
 */
        __pyx_t_3 = ((__pyx_v_endc == NULL) != 0);
        if (__pyx_t_3) {
        } else {
          __pyx_t_7 = __pyx_t_3;
          goto __pyx_L31_bool_binop_done;
        }
        __pyx_t_3 = ((__pyx_v_c != '=') != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L31_bool_binop_done:;
        if (unlikely(__pyx_t_7)) {

          /* "Orange/data/_io.pyx":152
 *                     raise ValueError("{}:{}:{}: quoted value should be "
 *                         "followed by value separator or end of line"
 *                         .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 # fall through
 *                 state = END_ATOM
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_quoted_value_should_be_followed, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 152, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 152, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L5_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_6 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 152, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_6);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_v_fname);
            __Pyx_GIVEREF(__pyx_v_fname);
            PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_v_fname);
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_9, __pyx_t_5);
            __pyx_t_8 = 0;
            __pyx_t_5 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Orange/data/_io.pyx":150
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":
 *                     raise ValueError("{}:{}:{}: quoted value should be "             # <<<<<<<<<<<<<<
 *                         "followed by value separator or end of line"
 *                         .format(fname, cur_line, col))
 */
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 150, __pyx_L5_error)

          /* "Orange/data/_io.pyx":149
 *                     continue
 *                 endc = strchr(not_in_atom, c)
 *                 if endc == NULL and c != b"=":             # <<<<<<<<<<<<<<
 *                     raise ValueError("{}:{}:{}: quoted value should be "
 *                         "followed by value separator or end of line"
 */
        }

        /* "Orange/data/_io.pyx":154
 *                         .format(fname, cur_line, col))
 *                 # fall through
 *                 state = END_ATOM             # <<<<<<<<<<<<<<
 * 
 *             if state == END_ATOM:
 */
        __pyx_v_state = __pyx_e_6Orange_4data_3_io_END_ATOM;

        /* "Orange/data/_io.pyx":145
 *                     continue
 * 
 *             if state == END_QUOTED:             # <<<<<<<<<<<<<<
 *                 if c == b" " or c == b"\t":
 *                     continue
 */
      }

      /* "Orange/data/_io.pyx":156
 *                 state = END_ATOM
 * 
 *             if state == END_ATOM:             # <<<<<<<<<<<<<<
 *                 while atomp != atom and (
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_END_ATOM) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":157
 * 
 *             if state == END_ATOM:
 *                 while atomp != atom and (             # <<<<<<<<<<<<<<
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):
 *                     atomp -= 1
 */
        while (1) {
          __pyx_t_3 = ((__pyx_v_atomp != __pyx_v_atom) != 0);
          if (__pyx_t_3) {
          } else {
            __pyx_t_7 = __pyx_t_3;
            goto __pyx_L36_bool_binop_done;
          }

          /* "Orange/data/_io.pyx":158
 *             if state == END_ATOM:
 *                 while atomp != atom and (
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):             # <<<<<<<<<<<<<<
 *                     atomp -= 1
 *                 if atomp == atom:
 */
          __pyx_t_3 = (((__pyx_v_atomp[-1L]) == ' ') != 0);
          if (!__pyx_t_3) {
          } else {
            __pyx_t_7 = __pyx_t_3;
            goto __pyx_L36_bool_binop_done;
          }
          __pyx_t_3 = (((__pyx_v_atomp[-1L]) == '\t') != 0);
          __pyx_t_7 = __pyx_t_3;
          __pyx_L36_bool_binop_done:;
          if (!__pyx_t_7) break;

          /* "Orange/data/_io.pyx":159
 *                 while atomp != atom and (
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):
 *                     atomp -= 1             # <<<<<<<<<<<<<<
 *                 if atomp == atom:
 *                     if c == b"=":
 */
          __pyx_v_atomp = (__pyx_v_atomp - 1);
        }

        /* "Orange/data/_io.pyx":160
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):
 *                     atomp -= 1
 *                 if atomp == atom:             # <<<<<<<<<<<<<<
 *                     if c == b"=":
 *                         raise ValueError("{}:{}:{}: empty value name"
 */
        __pyx_t_7 = ((__pyx_v_atomp == __pyx_v_atom) != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":161
 *                     atomp -= 1
 *                 if atomp == atom:
 *                     if c == b"=":             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{}: empty value name"
 *                             .format(fname, cur_line, col))
 */
          __pyx_t_7 = ((__pyx_v_c == '=') != 0);
          if (unlikely(__pyx_t_7)) {

            /* "Orange/data/_io.pyx":163
 *                     if c == b"=":
 *                         raise ValueError("{}:{}:{}: empty value name"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                     else:
 *                         state = TO_NEXT
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_empty_value_name, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 163, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 163, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_8 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 163, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_10);
              if (__pyx_t_8) {
                __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_6 = 0;
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "Orange/data/_io.pyx":162
 *                 if atomp == atom:
 *                     if c == b"=":
 *                         raise ValueError("{}:{}:{}: empty value name"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                     else:
 */
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(0, 162, __pyx_L5_error)

            /* "Orange/data/_io.pyx":161
 *                     atomp -= 1
 *                 if atomp == atom:
 *                     if c == b"=":             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{}: empty value name"
 *                             .format(fname, cur_line, col))
 */
          }

          /* "Orange/data/_io.pyx":165
 *                             .format(fname, cur_line, col))
 *                     else:
 *                         state = TO_NEXT             # <<<<<<<<<<<<<<
 *                 else:
 *                     atomp[0] = 0
 */
          /*else*/ {
            __pyx_v_state = __pyx_e_6Orange_4data_3_io_TO_NEXT;
          }

          /* "Orange/data/_io.pyx":160
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):
 *                     atomp -= 1
 *                 if atomp == atom:             # <<<<<<<<<<<<<<
 *                     if c == b"=":
 *                         raise ValueError("{}:{}:{}: empty value name"
 */
          goto __pyx_L39;
        }

        /* "Orange/data/_io.pyx":167
 *                         state = TO_NEXT
 *                 else:
 *                     atomp[0] = 0             # <<<<<<<<<<<<<<
 *                     b_atom = atom
 *                     """
 */
        /*else*/ {
          (__pyx_v_atomp[0]) = 0;

          /* "Orange/data/_io.pyx":168
 *                 else:
 *                     atomp[0] = 0
 *                     b_atom = atom             # <<<<<<<<<<<<<<
 *                     """
 *                     attr_index = t_names.get(b_atom, -1)
 */
          __pyx_v_b_atom = __pyx_v_atom;

          /* "Orange/data/_io.pyx":174
 *                         attr_index = t_names[b_atom] = len(t_names)
 *                     """
 *                     attr_index = t_names.setdefault(b_atom,len(t_names))             # <<<<<<<<<<<<<<
 * 
 *                     atomp = atom
 */
          if (unlikely(!__pyx_v_t_names)) { __Pyx_RaiseUnboundLocalError("t_names"); __PYX_ERR(0, 174, __pyx_L5_error) }
          __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_b_atom); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 174, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          if (unlikely(!__pyx_v_t_names)) { __Pyx_RaiseUnboundLocalError("t_names"); __PYX_ERR(0, 174, __pyx_L5_error) }
          __pyx_t_11 = PyDict_Size(__pyx_v_t_names); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 174, __pyx_L5_error)
          __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = __Pyx_PyDict_SetDefault(__pyx_v_t_names, __pyx_t_4, __pyx_t_1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 174, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_10); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 174, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_v_attr_index = __pyx_t_9;

          /* "Orange/data/_io.pyx":176
 *                     attr_index = t_names.setdefault(b_atom,len(t_names))
 * 
 *                     atomp = atom             # <<<<<<<<<<<<<<
 *                     if c == b"=":
 *                         value = 0
 */
          __pyx_v_atomp = __pyx_v_atom;

          /* "Orange/data/_io.pyx":177
 * 
 *                     atomp = atom
 *                     if c == b"=":             # <<<<<<<<<<<<<<
 *                         value = 0
 *                         state = WAIT_VALUE
 */
          __pyx_t_7 = ((__pyx_v_c == '=') != 0);
          if (__pyx_t_7) {

            /* "Orange/data/_io.pyx":178
 *                     atomp = atom
 *                     if c == b"=":
 *                         value = 0             # <<<<<<<<<<<<<<
 *                         state = WAIT_VALUE
 *                         continue
 */
            __pyx_v_value = 0.0;

            /* "Orange/data/_io.pyx":179
 *                     if c == b"=":
 *                         value = 0
 *                         state = WAIT_VALUE             # <<<<<<<<<<<<<<
 *                         continue
 *                     # fall through
 */
            __pyx_v_state = __pyx_e_6Orange_4data_3_io_WAIT_VALUE;

            /* "Orange/data/_io.pyx":180
 *                         value = 0
 *                         state = WAIT_VALUE
 *                         continue             # <<<<<<<<<<<<<<
 *                     # fall through
 *                     state = SET_VALUE
 */
            goto __pyx_L7_continue;

            /* "Orange/data/_io.pyx":177
 * 
 *                     atomp = atom
 *                     if c == b"=":             # <<<<<<<<<<<<<<
 *                         value = 0
 *                         state = WAIT_VALUE
 */
          }

          /* "Orange/data/_io.pyx":182
 *                         continue
 *                     # fall through
 *                     state = SET_VALUE             # <<<<<<<<<<<<<<
 * 
 *             if state == WAIT_VALUE:
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_SET_VALUE;
        }
        __pyx_L39:;

        /* "Orange/data/_io.pyx":156
 *                 state = END_ATOM
 * 
 *             if state == END_ATOM:             # <<<<<<<<<<<<<<
 *                 while atomp != atom and (
 *                         atomp[-1] == b" " or atomp[-1] == b"\t"):
 */
      }

      /* "Orange/data/_io.pyx":184
 *                     state = SET_VALUE
 * 
 *             if state == WAIT_VALUE:             # <<<<<<<<<<<<<<
 *                 if c == b" " or c == b"\t":
 *                     continue
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_WAIT_VALUE) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":185
 * 
 *             if state == WAIT_VALUE:
 *                 if c == b" " or c == b"\t":             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
        switch (__pyx_v_c) {
          case ' ':
          case '\t':

          /* "Orange/data/_io.pyx":186
 *             if state == WAIT_VALUE:
 *                 if c == b" " or c == b"\t":
 *                     continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     # fall through
 */
          goto __pyx_L7_continue;

          /* "Orange/data/_io.pyx":185
 * 
 *             if state == WAIT_VALUE:
 *                 if c == b" " or c == b"\t":             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
          break;
          default:

          /* "Orange/data/_io.pyx":189
 *                 else:
 *                     # fall through
 *                     state = READ_VALUE             # <<<<<<<<<<<<<<
 * 
 *             if state == READ_VALUE:
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_VALUE;
          break;
        }

        /* "Orange/data/_io.pyx":184
 *                     state = SET_VALUE
 * 
 *             if state == WAIT_VALUE:             # <<<<<<<<<<<<<<
 *                 if c == b" " or c == b"\t":
 *                     continue
 */
      }

      /* "Orange/data/_io.pyx":191
 *                     state = READ_VALUE
 * 
 *             if state == READ_VALUE:             # <<<<<<<<<<<<<<
 *                 if b"0" <= c <= b"9":
 *                     value = value * 10 + (c & 0xf)
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_READ_VALUE) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":192
 * 
 *             if state == READ_VALUE:
 *                 if b"0" <= c <= b"9":             # <<<<<<<<<<<<<<
 *                     value = value * 10 + (c & 0xf)
 *                 elif c == b".":
 */
        __pyx_t_7 = ('0' <= __pyx_v_c);
        if (__pyx_t_7) {
          __pyx_t_7 = (__pyx_v_c <= '9');
        }
        __pyx_t_3 = (__pyx_t_7 != 0);
        if (__pyx_t_3) {

          /* "Orange/data/_io.pyx":193
 *             if state == READ_VALUE:
 *                 if b"0" <= c <= b"9":
 *                     value = value * 10 + (c & 0xf)             # <<<<<<<<<<<<<<
 *                 elif c == b".":
 *                     decs = 0.1
 */
          __pyx_v_value = ((__pyx_v_value * 10.0) + (__pyx_v_c & 0xf));

          /* "Orange/data/_io.pyx":192
 * 
 *             if state == READ_VALUE:
 *                 if b"0" <= c <= b"9":             # <<<<<<<<<<<<<<
 *                     value = value * 10 + (c & 0xf)
 *                 elif c == b".":
 */
          goto __pyx_L44;
        }

        /* "Orange/data/_io.pyx":194
 *                 if b"0" <= c <= b"9":
 *                     value = value * 10 + (c & 0xf)
 *                 elif c == b".":             # <<<<<<<<<<<<<<
 *                     decs = 0.1
 *                     state = READ_DECS
 */
        __pyx_t_3 = ((__pyx_v_c == '.') != 0);
        if (__pyx_t_3) {

          /* "Orange/data/_io.pyx":195
 *                     value = value * 10 + (c & 0xf)
 *                 elif c == b".":
 *                     decs = 0.1             # <<<<<<<<<<<<<<
 *                     state = READ_DECS
 *                 else:
 */
          __pyx_v_decs = 0.1;

          /* "Orange/data/_io.pyx":196
 *                 elif c == b".":
 *                     decs = 0.1
 *                     state = READ_DECS             # <<<<<<<<<<<<<<
 *                 else:
 *                     endc = strchr(not_in_atom, c)
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_DECS;

          /* "Orange/data/_io.pyx":194
 *                 if b"0" <= c <= b"9":
 *                     value = value * 10 + (c & 0xf)
 *                 elif c == b".":             # <<<<<<<<<<<<<<
 *                     decs = 0.1
 *                     state = READ_DECS
 */
          goto __pyx_L44;
        }

        /* "Orange/data/_io.pyx":198
 *                     state = READ_DECS
 *                 else:
 *                     endc = strchr(not_in_atom, c)             # <<<<<<<<<<<<<<
 *                     if endc != NULL:
 *                         state = SET_VALUE
 */
        /*else*/ {
          __pyx_v_endc = strchr(__pyx_v_not_in_atom, __pyx_v_c);

          /* "Orange/data/_io.pyx":199
 *                 else:
 *                     endc = strchr(not_in_atom, c)
 *                     if endc != NULL:             # <<<<<<<<<<<<<<
 *                         state = SET_VALUE
 *                     else:
 */
          __pyx_t_3 = ((__pyx_v_endc != NULL) != 0);
          if (likely(__pyx_t_3)) {

            /* "Orange/data/_io.pyx":200
 *                     endc = strchr(not_in_atom, c)
 *                     if endc != NULL:
 *                         state = SET_VALUE             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"
 */
            __pyx_v_state = __pyx_e_6Orange_4data_3_io_SET_VALUE;

            /* "Orange/data/_io.pyx":199
 *                 else:
 *                     endc = strchr(not_in_atom, c)
 *                     if endc != NULL:             # <<<<<<<<<<<<<<
 *                         state = SET_VALUE
 *                     else:
 */
            goto __pyx_L45;
          }

          /* "Orange/data/_io.pyx":202
 *                         state = SET_VALUE
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                 continue
 */
          /*else*/ {

            /* "Orange/data/_io.pyx":203
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_value, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 203, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 203, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_4, __pyx_t_5};
              __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 203, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_4, __pyx_t_5};
              __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 203, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 203, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (__pyx_t_6) {
                __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_4 = 0;
              __pyx_t_5 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 203, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "Orange/data/_io.pyx":202
 *                         state = SET_VALUE
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                 continue
 */
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 202, __pyx_L5_error)
          }
          __pyx_L45:;
        }
        __pyx_L44:;

        /* "Orange/data/_io.pyx":204
 *                         raise ValueError("{}:{}:{}: invalid value"
 *                             .format(fname, cur_line, col))
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             if state == READ_DECS:
 */
        goto __pyx_L7_continue;

        /* "Orange/data/_io.pyx":191
 *                     state = READ_VALUE
 * 
 *             if state == READ_VALUE:             # <<<<<<<<<<<<<<
 *                 if b"0" <= c <= b"9":
 *                     value = value * 10 + (c & 0xf)
 */
      }

      /* "Orange/data/_io.pyx":206
 *                 continue
 * 
 *             if state == READ_DECS:             # <<<<<<<<<<<<<<
 *                 if b"0" <= c <= b"9":
 *                     value = value * decs + (c & 0xf)
 */
      __pyx_t_3 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_READ_DECS) != 0);
      if (__pyx_t_3) {

        /* "Orange/data/_io.pyx":207
 * 
 *             if state == READ_DECS:
 *                 if b"0" <= c <= b"9":             # <<<<<<<<<<<<<<
 *                     value = value * decs + (c & 0xf)
 *                     decs /= 10
 */
        __pyx_t_3 = ('0' <= __pyx_v_c);
        if (__pyx_t_3) {
          __pyx_t_3 = (__pyx_v_c <= '9');
        }
        __pyx_t_7 = (__pyx_t_3 != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":208
 *             if state == READ_DECS:
 *                 if b"0" <= c <= b"9":
 *                     value = value * decs + (c & 0xf)             # <<<<<<<<<<<<<<
 *                     decs /= 10
 *                 else:
 */
          __pyx_v_value = ((__pyx_v_value * __pyx_v_decs) + (__pyx_v_c & 0xf));

          /* "Orange/data/_io.pyx":209
 *                 if b"0" <= c <= b"9":
 *                     value = value * decs + (c & 0xf)
 *                     decs /= 10             # <<<<<<<<<<<<<<
 *                 else:
 *                     endc = strchr(not_in_atom, c)
 */
          __pyx_v_decs = (__pyx_v_decs / 10.0);

          /* "Orange/data/_io.pyx":207
 * 
 *             if state == READ_DECS:
 *                 if b"0" <= c <= b"9":             # <<<<<<<<<<<<<<
 *                     value = value * decs + (c & 0xf)
 *                     decs /= 10
 */
          goto __pyx_L47;
        }

        /* "Orange/data/_io.pyx":211
 *                     decs /= 10
 *                 else:
 *                     endc = strchr(not_in_atom, c)             # <<<<<<<<<<<<<<
 *                     if endc != NULL:
 *                         state = SET_VALUE
 */
        /*else*/ {
          __pyx_v_endc = strchr(__pyx_v_not_in_atom, __pyx_v_c);

          /* "Orange/data/_io.pyx":212
 *                 else:
 *                     endc = strchr(not_in_atom, c)
 *                     if endc != NULL:             # <<<<<<<<<<<<<<
 *                         state = SET_VALUE
 *                     else:
 */
          __pyx_t_7 = ((__pyx_v_endc != NULL) != 0);
          if (likely(__pyx_t_7)) {

            /* "Orange/data/_io.pyx":213
 *                     endc = strchr(not_in_atom, c)
 *                     if endc != NULL:
 *                         state = SET_VALUE             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"
 */
            __pyx_v_state = __pyx_e_6Orange_4data_3_io_SET_VALUE;

            /* "Orange/data/_io.pyx":212
 *                 else:
 *                     endc = strchr(not_in_atom, c)
 *                     if endc != NULL:             # <<<<<<<<<<<<<<
 *                         state = SET_VALUE
 *                     else:
 */
            goto __pyx_L48;
          }

          /* "Orange/data/_io.pyx":215
 *                         state = SET_VALUE
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                 continue
 */
          /*else*/ {

            /* "Orange/data/_io.pyx":216
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_value, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 216, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 216, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 216, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_4 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_fname, __pyx_t_8, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_6 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (__pyx_t_4) {
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_8);
              PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_8 = 0;
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "Orange/data/_io.pyx":215
 *                         state = SET_VALUE
 *                     else:
 *                         raise ValueError("{}:{}:{}: invalid value"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                 continue
 */
            __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 215, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_Raise(__pyx_t_10, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __PYX_ERR(0, 215, __pyx_L5_error)
          }
          __pyx_L48:;
        }
        __pyx_L47:;

        /* "Orange/data/_io.pyx":217
 *                         raise ValueError("{}:{}:{}: invalid value"
 *                             .format(fname, cur_line, col))
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             if state == SET_VALUE:
 */
        goto __pyx_L7_continue;

        /* "Orange/data/_io.pyx":206
 *                 continue
 * 
 *             if state == READ_DECS:             # <<<<<<<<<<<<<<
 *                 if b"0" <= c <= b"9":
 *                     value = value * decs + (c & 0xf)
 */
      }

      /* "Orange/data/_io.pyx":219
 *                 continue
 * 
 *             if state == SET_VALUE:             # <<<<<<<<<<<<<<
 *                 if col_kind == ATTRIBUTE:
 *                     X_data.append(value)
 */
      __pyx_t_7 = ((__pyx_v_state == __pyx_e_6Orange_4data_3_io_SET_VALUE) != 0);
      if (__pyx_t_7) {

        /* "Orange/data/_io.pyx":220
 * 
 *             if state == SET_VALUE:
 *                 if col_kind == ATTRIBUTE:             # <<<<<<<<<<<<<<
 *                     X_data.append(value)
 *                     X_rows.append(line - 1)
 */
        switch (__pyx_v_col_kind) {
          case __pyx_e_6Orange_4data_3_io_ATTRIBUTE:

          /* "Orange/data/_io.pyx":221
 *             if state == SET_VALUE:
 *                 if col_kind == ATTRIBUTE:
 *                     X_data.append(value)             # <<<<<<<<<<<<<<
 *                     X_rows.append(line - 1)
 *                     X_cols.append(attr_index)
 */
          __pyx_t_10 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 221, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_X_data, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 221, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":222
 *                 if col_kind == ATTRIBUTE:
 *                     X_data.append(value)
 *                     X_rows.append(line - 1)             # <<<<<<<<<<<<<<
 *                     X_cols.append(attr_index)
 *                 elif col_kind == CLASS:
 */
          __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_line - 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 222, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_X_rows, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 222, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":223
 *                     X_data.append(value)
 *                     X_rows.append(line - 1)
 *                     X_cols.append(attr_index)             # <<<<<<<<<<<<<<
 *                 elif col_kind == CLASS:
 *                     Y_data.append(value)
 */
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_attr_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 223, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_X_cols, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 223, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":220
 * 
 *             if state == SET_VALUE:
 *                 if col_kind == ATTRIBUTE:             # <<<<<<<<<<<<<<
 *                     X_data.append(value)
 *                     X_rows.append(line - 1)
 */
          break;

          /* "Orange/data/_io.pyx":224
 *                     X_rows.append(line - 1)
 *                     X_cols.append(attr_index)
 *                 elif col_kind == CLASS:             # <<<<<<<<<<<<<<
 *                     Y_data.append(value)
 *                     Y_rows.append(line - 1)
 */
          case __pyx_e_6Orange_4data_3_io_CLASS:

          /* "Orange/data/_io.pyx":225
 *                     X_cols.append(attr_index)
 *                 elif col_kind == CLASS:
 *                     Y_data.append(value)             # <<<<<<<<<<<<<<
 *                     Y_rows.append(line - 1)
 *                     Y_cols.append(attr_index)
 */
          __pyx_t_10 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 225, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_Y_data, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 225, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":226
 *                 elif col_kind == CLASS:
 *                     Y_data.append(value)
 *                     Y_rows.append(line - 1)             # <<<<<<<<<<<<<<
 *                     Y_cols.append(attr_index)
 *                 else:
 */
          __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_line - 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 226, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_Y_rows, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 226, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":227
 *                     Y_data.append(value)
 *                     Y_rows.append(line - 1)
 *                     Y_cols.append(attr_index)             # <<<<<<<<<<<<<<
 *                 else:
 *                     M_data.append(value)
 */
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_attr_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 227, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_Y_cols, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 227, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":224
 *                     X_rows.append(line - 1)
 *                     X_cols.append(attr_index)
 *                 elif col_kind == CLASS:             # <<<<<<<<<<<<<<
 *                     Y_data.append(value)
 *                     Y_rows.append(line - 1)
 */
          break;
          default:

          /* "Orange/data/_io.pyx":229
 *                     Y_cols.append(attr_index)
 *                 else:
 *                     M_data.append(value)             # <<<<<<<<<<<<<<
 *                     M_rows.append(line - 1)
 *                     M_cols.append(attr_index)
 */
          __pyx_t_10 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 229, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_M_data, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 229, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":230
 *                 else:
 *                     M_data.append(value)
 *                     M_rows.append(line - 1)             # <<<<<<<<<<<<<<
 *                     M_cols.append(attr_index)
 *                 in_line += 1
 */
          __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_line - 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 230, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_M_rows, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 230, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Orange/data/_io.pyx":231
 *                     M_data.append(value)
 *                     M_rows.append(line - 1)
 *                     M_cols.append(attr_index)             # <<<<<<<<<<<<<<
 *                 in_line += 1
 *                 state = TO_NEXT
 */
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_attr_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 231, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_M_cols, __pyx_t_10); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 231, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          break;
        }

        /* "Orange/data/_io.pyx":232
 *                     M_rows.append(line - 1)
 *                     M_cols.append(attr_index)
 *                 in_line += 1             # <<<<<<<<<<<<<<
 *                 state = TO_NEXT
 * 
 */
        __pyx_v_in_line = (__pyx_v_in_line + 1);

        /* "Orange/data/_io.pyx":233
 *                     M_cols.append(attr_index)
 *                 in_line += 1
 *                 state = TO_NEXT             # <<<<<<<<<<<<<<
 * 
 *             if state == TO_NEXT:
 */
        __pyx_v_state = __pyx_e_6Orange_4data_3_io_TO_NEXT;

        /* "Orange/data/_io.pyx":219
 *                 continue
 * 
 *             if state == SET_VALUE:             # <<<<<<<<<<<<<<
 *                 if col_kind == ATTRIBUTE:
 *                     X_data.append(value)
 */
      }

      /* "Orange/data/_io.pyx":235
 *                 state = TO_NEXT
 * 
 *             if state == TO_NEXT:             # <<<<<<<<<<<<<<
 *                 if c == b"|":
 *                     if col_kind != ATTRIBUTE:
 */
      switch (__pyx_v_state) {
        case __pyx_e_6Orange_4data_3_io_TO_NEXT:

        /* "Orange/data/_io.pyx":236
 * 
 *             if state == TO_NEXT:
 *                 if c == b"|":             # <<<<<<<<<<<<<<
 *                     if col_kind != ATTRIBUTE:
 *                         raise ValueError(
 */
        __pyx_t_7 = ((__pyx_v_c == '|') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":237
 *             if state == TO_NEXT:
 *                 if c == b"|":
 *                     if col_kind != ATTRIBUTE:             # <<<<<<<<<<<<<<
 *                         raise ValueError(
 *                             "{}:{}:{}: classes should follow attributes"
 */
          __pyx_t_7 = ((__pyx_v_col_kind != __pyx_e_6Orange_4data_3_io_ATTRIBUTE) != 0);
          if (unlikely(__pyx_t_7)) {

            /* "Orange/data/_io.pyx":240
 *                         raise ValueError(
 *                             "{}:{}:{}: classes should follow attributes"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                     col_kind = CLASS
 *                     t_names = class_indices
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_classes_should_follow_attribute, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 240, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 240, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_8 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
              __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 240, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_fname, __pyx_t_6, __pyx_t_5};
              __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 240, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 240, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_4);
              if (__pyx_t_8) {
                __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_6 = 0;
              __pyx_t_5 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 240, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "Orange/data/_io.pyx":238
 *                 if c == b"|":
 *                     if col_kind != ATTRIBUTE:
 *                         raise ValueError(             # <<<<<<<<<<<<<<
 *                             "{}:{}:{}: classes should follow attributes"
 *                             .format(fname, cur_line, col))
 */
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 238, __pyx_L5_error)

            /* "Orange/data/_io.pyx":237
 *             if state == TO_NEXT:
 *                 if c == b"|":
 *                     if col_kind != ATTRIBUTE:             # <<<<<<<<<<<<<<
 *                         raise ValueError(
 *                             "{}:{}:{}: classes should follow attributes"
 */
          }

          /* "Orange/data/_io.pyx":241
 *                             "{}:{}:{}: classes should follow attributes"
 *                             .format(fname, cur_line, col))
 *                     col_kind = CLASS             # <<<<<<<<<<<<<<
 *                     t_names = class_indices
 *                     state = READ_START_ATOM
 */
          __pyx_v_col_kind = __pyx_e_6Orange_4data_3_io_CLASS;

          /* "Orange/data/_io.pyx":242
 *                             .format(fname, cur_line, col))
 *                     col_kind = CLASS
 *                     t_names = class_indices             # <<<<<<<<<<<<<<
 *                     state = READ_START_ATOM
 *                 elif c == b";":
 */
          __Pyx_INCREF(__pyx_v_class_indices);
          __Pyx_XDECREF_SET(__pyx_v_t_names, __pyx_v_class_indices);

          /* "Orange/data/_io.pyx":243
 *                     col_kind = CLASS
 *                     t_names = class_indices
 *                     state = READ_START_ATOM             # <<<<<<<<<<<<<<
 *                 elif c == b";":
 *                     if col_kind == META:
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_START_ATOM;

          /* "Orange/data/_io.pyx":236
 * 
 *             if state == TO_NEXT:
 *                 if c == b"|":             # <<<<<<<<<<<<<<
 *                     if col_kind != ATTRIBUTE:
 *                         raise ValueError(
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":244
 *                     t_names = class_indices
 *                     state = READ_START_ATOM
 *                 elif c == b";":             # <<<<<<<<<<<<<<
 *                     if col_kind == META:
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"
 */
        __pyx_t_7 = ((__pyx_v_c == ';') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":245
 *                     state = READ_START_ATOM
 *                 elif c == b";":
 *                     if col_kind == META:             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"
 *                             .format(fname, cur_line, col))
 */
          __pyx_t_7 = ((__pyx_v_col_kind == __pyx_e_6Orange_4data_3_io_META) != 0);
          if (unlikely(__pyx_t_7)) {

            /* "Orange/data/_io.pyx":247
 *                     if col_kind == META:
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"
 *                             .format(fname, cur_line, col))             # <<<<<<<<<<<<<<
 *                     col_kind = META
 *                     t_names = meta_indices
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_duplicated_semi_colons, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 247, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_cur_line); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 247, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_col); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 247, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_4, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_fname, __pyx_t_4, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L5_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 247, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (__pyx_t_6) {
                __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
              }
              __Pyx_INCREF(__pyx_v_fname);
              __Pyx_GIVEREF(__pyx_v_fname);
              PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_fname);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_t_5);
              __pyx_t_4 = 0;
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L5_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "Orange/data/_io.pyx":246
 *                 elif c == b";":
 *                     if col_kind == META:
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"             # <<<<<<<<<<<<<<
 *                             .format(fname, cur_line, col))
 *                     col_kind = META
 */
            __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 246, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_Raise(__pyx_t_10, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __PYX_ERR(0, 246, __pyx_L5_error)

            /* "Orange/data/_io.pyx":245
 *                     state = READ_START_ATOM
 *                 elif c == b";":
 *                     if col_kind == META:             # <<<<<<<<<<<<<<
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"
 *                             .format(fname, cur_line, col))
 */
          }

          /* "Orange/data/_io.pyx":248
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"
 *                             .format(fname, cur_line, col))
 *                     col_kind = META             # <<<<<<<<<<<<<<
 *                     t_names = meta_indices
 *                     state = READ_START_ATOM
 */
          __pyx_v_col_kind = __pyx_e_6Orange_4data_3_io_META;

          /* "Orange/data/_io.pyx":249
 *                             .format(fname, cur_line, col))
 *                     col_kind = META
 *                     t_names = meta_indices             # <<<<<<<<<<<<<<
 *                     state = READ_START_ATOM
 *                 elif c == b",":
 */
          __Pyx_INCREF(__pyx_v_meta_indices);
          __Pyx_XDECREF_SET(__pyx_v_t_names, __pyx_v_meta_indices);

          /* "Orange/data/_io.pyx":250
 *                     col_kind = META
 *                     t_names = meta_indices
 *                     state = READ_START_ATOM             # <<<<<<<<<<<<<<
 *                 elif c == b",":
 *                     state = READ_START_ATOM
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_START_ATOM;

          /* "Orange/data/_io.pyx":244
 *                     t_names = class_indices
 *                     state = READ_START_ATOM
 *                 elif c == b";":             # <<<<<<<<<<<<<<
 *                     if col_kind == META:
 *                         raise ValueError("{}:{}:{} duplicated semi-colons"
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":251
 *                     t_names = meta_indices
 *                     state = READ_START_ATOM
 *                 elif c == b",":             # <<<<<<<<<<<<<<
 *                     state = READ_START_ATOM
 *                 elif c == b"#":
 */
        __pyx_t_7 = ((__pyx_v_c == ',') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":252
 *                     state = READ_START_ATOM
 *                 elif c == b",":
 *                     state = READ_START_ATOM             # <<<<<<<<<<<<<<
 *                 elif c == b"#":
 *                     state = COMMENT
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_START_ATOM;

          /* "Orange/data/_io.pyx":251
 *                     t_names = meta_indices
 *                     state = READ_START_ATOM
 *                 elif c == b",":             # <<<<<<<<<<<<<<
 *                     state = READ_START_ATOM
 *                 elif c == b"#":
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":253
 *                 elif c == b",":
 *                     state = READ_START_ATOM
 *                 elif c == b"#":             # <<<<<<<<<<<<<<
 *                     state = COMMENT
 *                 elif c == b"\\":
 */
        __pyx_t_7 = ((__pyx_v_c == '#') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":254
 *                     state = READ_START_ATOM
 *                 elif c == b"#":
 *                     state = COMMENT             # <<<<<<<<<<<<<<
 *                 elif c == b"\\":
 *                     state = ESCAPE
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_COMMENT;

          /* "Orange/data/_io.pyx":253
 *                 elif c == b",":
 *                     state = READ_START_ATOM
 *                 elif c == b"#":             # <<<<<<<<<<<<<<
 *                     state = COMMENT
 *                 elif c == b"\\":
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":255
 *                 elif c == b"#":
 *                     state = COMMENT
 *                 elif c == b"\\":             # <<<<<<<<<<<<<<
 *                     state = ESCAPE
 *                 elif c == b"\n" or f_eof:
 */
        __pyx_t_7 = ((__pyx_v_c == '\\') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":256
 *                     state = COMMENT
 *                 elif c == b"\\":
 *                     state = ESCAPE             # <<<<<<<<<<<<<<
 *                 elif c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_ESCAPE;

          /* "Orange/data/_io.pyx":255
 *                 elif c == b"#":
 *                     state = COMMENT
 *                 elif c == b"\\":             # <<<<<<<<<<<<<<
 *                     state = ESCAPE
 *                 elif c == b"\n" or f_eof:
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":257
 *                 elif c == b"\\":
 *                     state = ESCAPE
 *                 elif c == b"\n" or f_eof:             # <<<<<<<<<<<<<<
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":
 */
        __pyx_t_3 = ((__pyx_v_c == '\n') != 0);
        if (!__pyx_t_3) {
        } else {
          __pyx_t_7 = __pyx_t_3;
          goto __pyx_L53_bool_binop_done;
        }
        __pyx_t_3 = (__pyx_v_f_eof != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L53_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":258
 *                     state = ESCAPE
 *                 elif c == b"\n" or f_eof:
 *                     state = BEGIN_LINE             # <<<<<<<<<<<<<<
 *                 elif c == b"\r":
 *                     state = CARRIAGE_RETURNED
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_BEGIN_LINE;

          /* "Orange/data/_io.pyx":257
 *                 elif c == b"\\":
 *                     state = ESCAPE
 *                 elif c == b"\n" or f_eof:             # <<<<<<<<<<<<<<
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":259
 *                 elif c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":             # <<<<<<<<<<<<<<
 *                     state = CARRIAGE_RETURNED
 *                 else:
 */
        __pyx_t_7 = ((__pyx_v_c == '\r') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":260
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":
 *                     state = CARRIAGE_RETURNED             # <<<<<<<<<<<<<<
 *                 else:
 *                     state = READ_START_ATOM
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_CARRIAGE_RETURNED;

          /* "Orange/data/_io.pyx":259
 *                 elif c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":             # <<<<<<<<<<<<<<
 *                     state = CARRIAGE_RETURNED
 *                 else:
 */
          goto __pyx_L50;
        }

        /* "Orange/data/_io.pyx":262
 *                     state = CARRIAGE_RETURNED
 *                 else:
 *                     state = READ_START_ATOM             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        /*else*/ {
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_READ_START_ATOM;
        }
        __pyx_L50:;

        /* "Orange/data/_io.pyx":263
 *                 else:
 *                     state = READ_START_ATOM
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             elif state == COMMENT:
 */
        goto __pyx_L7_continue;

        /* "Orange/data/_io.pyx":235
 *                 state = TO_NEXT
 * 
 *             if state == TO_NEXT:             # <<<<<<<<<<<<<<
 *                 if c == b"|":
 *                     if col_kind != ATTRIBUTE:
 */
        break;

        /* "Orange/data/_io.pyx":265
 *                 continue
 * 
 *             elif state == COMMENT:             # <<<<<<<<<<<<<<
 *                 if c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 */
        case __pyx_e_6Orange_4data_3_io_COMMENT:

        /* "Orange/data/_io.pyx":266
 * 
 *             elif state == COMMENT:
 *                 if c == b"\n" or f_eof:             # <<<<<<<<<<<<<<
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":
 */
        __pyx_t_3 = ((__pyx_v_c == '\n') != 0);
        if (!__pyx_t_3) {
        } else {
          __pyx_t_7 = __pyx_t_3;
          goto __pyx_L56_bool_binop_done;
        }
        __pyx_t_3 = (__pyx_v_f_eof != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L56_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":267
 *             elif state == COMMENT:
 *                 if c == b"\n" or f_eof:
 *                     state = BEGIN_LINE             # <<<<<<<<<<<<<<
 *                 elif c == b"\r":
 *                     state = CARRIAGE_RETURNED
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_BEGIN_LINE;

          /* "Orange/data/_io.pyx":266
 * 
 *             elif state == COMMENT:
 *                 if c == b"\n" or f_eof:             # <<<<<<<<<<<<<<
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":
 */
          goto __pyx_L55;
        }

        /* "Orange/data/_io.pyx":268
 *                 if c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":             # <<<<<<<<<<<<<<
 *                     state = CARRIAGE_RETURNED
 * 
 */
        __pyx_t_7 = ((__pyx_v_c == '\r') != 0);
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":269
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":
 *                     state = CARRIAGE_RETURNED             # <<<<<<<<<<<<<<
 * 
 *             elif state == CARRIAGE_RETURNED:
 */
          __pyx_v_state = __pyx_e_6Orange_4data_3_io_CARRIAGE_RETURNED;

          /* "Orange/data/_io.pyx":268
 *                 if c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 *                 elif c == b"\r":             # <<<<<<<<<<<<<<
 *                     state = CARRIAGE_RETURNED
 * 
 */
        }
        __pyx_L55:;

        /* "Orange/data/_io.pyx":265
 *                 continue
 * 
 *             elif state == COMMENT:             # <<<<<<<<<<<<<<
 *                 if c == b"\n" or f_eof:
 *                     state = BEGIN_LINE
 */
        break;

        /* "Orange/data/_io.pyx":271
 *                     state = CARRIAGE_RETURNED
 * 
 *             elif state == CARRIAGE_RETURNED:             # <<<<<<<<<<<<<<
 *                 if c != b"\n" and not f_eof:
 *                     ungetc(c, f)
 */
        case __pyx_e_6Orange_4data_3_io_CARRIAGE_RETURNED:

        /* "Orange/data/_io.pyx":272
 * 
 *             elif state == CARRIAGE_RETURNED:
 *                 if c != b"\n" and not f_eof:             # <<<<<<<<<<<<<<
 *                     ungetc(c, f)
 *                 state = BEGIN_LINE
 */
        __pyx_t_3 = ((__pyx_v_c != '\n') != 0);
        if (__pyx_t_3) {
        } else {
          __pyx_t_7 = __pyx_t_3;
          goto __pyx_L59_bool_binop_done;
        }
        __pyx_t_3 = ((!(__pyx_v_f_eof != 0)) != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L59_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Orange/data/_io.pyx":273
 *             elif state == CARRIAGE_RETURNED:
 *                 if c != b"\n" and not f_eof:
 *                     ungetc(c, f)             # <<<<<<<<<<<<<<
 *                 state = BEGIN_LINE
 *     finally:
 */
          (void)(ungetc(__pyx_v_c, __pyx_v_f));

          /* "Orange/data/_io.pyx":272
 * 
 *             elif state == CARRIAGE_RETURNED:
 *                 if c != b"\n" and not f_eof:             # <<<<<<<<<<<<<<
 *                     ungetc(c, f)
 *                 state = BEGIN_LINE
 */
        }

        /* "Orange/data/_io.pyx":274
 *                 if c != b"\n" and not f_eof:
 *                     ungetc(c, f)
 *                 state = BEGIN_LINE             # <<<<<<<<<<<<<<
 *     finally:
 *         fclose(f)
 */
        __pyx_v_state = __pyx_e_6Orange_4data_3_io_BEGIN_LINE;

        /* "Orange/data/_io.pyx":271
 *                     state = CARRIAGE_RETURNED
 * 
 *             elif state == CARRIAGE_RETURNED:             # <<<<<<<<<<<<<<
 *                 if c != b"\n" and not f_eof:
 *                     ungetc(c, f)
 */
        break;
        default: break;
      }
      __pyx_L7_continue:;
    }
  }

  /* "Orange/data/_io.pyx":276
 *                 state = BEGIN_LINE
 *     finally:
 *         fclose(f)             # <<<<<<<<<<<<<<
 * 
 *     if in_line == 0:
 */
  /*finally:*/ {
    /*normal exit:*/{
      (void)(fclose(__pyx_v_f));
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __pyx_t_9 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
      {
        (void)(fclose(__pyx_v_f));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "Orange/data/_io.pyx":278
 *         fclose(f)
 * 
 *     if in_line == 0:             # <<<<<<<<<<<<<<
 *         line -= 1
 * 
 */
  __pyx_t_7 = ((__pyx_v_in_line == 0) != 0);
  if (__pyx_t_7) {

    /* "Orange/data/_io.pyx":279
 * 
 *     if in_line == 0:
 *         line -= 1             # <<<<<<<<<<<<<<
 * 
 *     res = []
 */
    __pyx_v_line = (__pyx_v_line - 1);

    /* "Orange/data/_io.pyx":278
 *         fclose(f)
 * 
 *     if in_line == 0:             # <<<<<<<<<<<<<<
 *         line -= 1
 * 
 */
  }

  /* "Orange/data/_io.pyx":281
 *         line -= 1
 * 
 *     res = []             # <<<<<<<<<<<<<<
 *     for t_data, t_rows, t_cols, ll in (
 *              (X_data, X_rows, X_cols, attr_indices),
 */
  __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_res = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "Orange/data/_io.pyx":283
 *     res = []
 *     for t_data, t_rows, t_cols, ll in (
 *              (X_data, X_rows, X_cols, attr_indices),             # <<<<<<<<<<<<<<
 *              (Y_data, Y_rows, Y_cols, class_indices),
 *              (M_data, M_rows, M_cols, meta_indices)):
 */
  __pyx_t_10 = PyTuple_New(4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_X_data);
  __Pyx_GIVEREF(__pyx_v_X_data);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_X_data);
  __Pyx_INCREF(__pyx_v_X_rows);
  __Pyx_GIVEREF(__pyx_v_X_rows);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_X_rows);
  __Pyx_INCREF(__pyx_v_X_cols);
  __Pyx_GIVEREF(__pyx_v_X_cols);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_X_cols);
  __Pyx_INCREF(__pyx_v_attr_indices);
  __Pyx_GIVEREF(__pyx_v_attr_indices);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_v_attr_indices);

  /* "Orange/data/_io.pyx":284
 *     for t_data, t_rows, t_cols, ll in (
 *              (X_data, X_rows, X_cols, attr_indices),
 *              (Y_data, Y_rows, Y_cols, class_indices),             # <<<<<<<<<<<<<<
 *              (M_data, M_rows, M_cols, meta_indices)):
 *         if len(t_data):
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_Y_data);
  __Pyx_GIVEREF(__pyx_v_Y_data);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_Y_data);
  __Pyx_INCREF(__pyx_v_Y_rows);
  __Pyx_GIVEREF(__pyx_v_Y_rows);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_Y_rows);
  __Pyx_INCREF(__pyx_v_Y_cols);
  __Pyx_GIVEREF(__pyx_v_Y_cols);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_Y_cols);
  __Pyx_INCREF(__pyx_v_class_indices);
  __Pyx_GIVEREF(__pyx_v_class_indices);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_class_indices);

  /* "Orange/data/_io.pyx":285
 *              (X_data, X_rows, X_cols, attr_indices),
 *              (Y_data, Y_rows, Y_cols, class_indices),
 *              (M_data, M_rows, M_cols, meta_indices)):             # <<<<<<<<<<<<<<
 *         if len(t_data):
 *             mat = sp.coo_matrix((t_data, (t_rows, t_cols)), (line, len(ll))).tocsr()
 */
  __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_M_data);
  __Pyx_GIVEREF(__pyx_v_M_data);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_M_data);
  __Pyx_INCREF(__pyx_v_M_rows);
  __Pyx_GIVEREF(__pyx_v_M_rows);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_M_rows);
  __Pyx_INCREF(__pyx_v_M_cols);
  __Pyx_GIVEREF(__pyx_v_M_cols);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_M_cols);
  __Pyx_INCREF(__pyx_v_meta_indices);
  __Pyx_GIVEREF(__pyx_v_meta_indices);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_v_meta_indices);

  /* "Orange/data/_io.pyx":283
 *     res = []
 *     for t_data, t_rows, t_cols, ll in (
 *              (X_data, X_rows, X_cols, attr_indices),             # <<<<<<<<<<<<<<
 *              (Y_data, Y_rows, Y_cols, class_indices),
 *              (M_data, M_rows, M_cols, meta_indices)):
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_8);
  __pyx_t_10 = 0;
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;

  /* "Orange/data/_io.pyx":282
 * 
 *     res = []
 *     for t_data, t_rows, t_cols, ll in (             # <<<<<<<<<<<<<<
 *              (X_data, X_rows, X_cols, attr_indices),
 *              (Y_data, Y_rows, Y_cols, class_indices),
 */
  __pyx_t_8 = __pyx_t_5; __Pyx_INCREF(__pyx_t_8); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_11 >= 3) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_5); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 282, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    if (likely(__pyx_t_5 != Py_None)) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 282, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 3); 
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_1,&__pyx_t_10,&__pyx_t_4,&__pyx_t_6};
        for (i=0; i < 4; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 282, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 282, __pyx_L1_error)
    }
    __Pyx_XDECREF_SET(__pyx_v_t_data, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_t_rows, __pyx_t_10);
    __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_t_cols, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_ll, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Orange/data/_io.pyx":286
 *              (Y_data, Y_rows, Y_cols, class_indices),
 *              (M_data, M_rows, M_cols, meta_indices)):
 *         if len(t_data):             # <<<<<<<<<<<<<<
 *             mat = sp.coo_matrix((t_data, (t_rows, t_cols)), (line, len(ll))).tocsr()
 *             mat.sort_indices()
 */
    __pyx_t_21 = PyObject_Length(__pyx_v_t_data); if (unlikely(__pyx_t_21 == ((Py_ssize_t)-1))) __PYX_ERR(0, 286, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_21 != 0);
    if (__pyx_t_7) {

      /* "Orange/data/_io.pyx":287
 *              (M_data, M_rows, M_cols, meta_indices)):
 *         if len(t_data):
 *             mat = sp.coo_matrix((t_data, (t_rows, t_cols)), (line, len(ll))).tocsr()             # <<<<<<<<<<<<<<
 *             mat.sort_indices()
 *         else:
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_sp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_coo_matrix); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_t_rows);
      __Pyx_GIVEREF(__pyx_v_t_rows);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_t_rows);
      __Pyx_INCREF(__pyx_v_t_cols);
      __Pyx_GIVEREF(__pyx_v_t_cols);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_t_cols);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_t_data);
      __Pyx_GIVEREF(__pyx_v_t_data);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_t_data);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_line); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_21 = PyObject_Length(__pyx_v_ll); if (unlikely(__pyx_t_21 == ((Py_ssize_t)-1))) __PYX_ERR(0, 287, __pyx_L1_error)
      __pyx_t_22 = PyInt_FromSsize_t(__pyx_t_21); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_22);
      PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_22);
      __pyx_t_4 = 0;
      __pyx_t_22 = 0;
      __pyx_t_22 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_22)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_22);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_22, __pyx_t_1, __pyx_t_23};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_22, __pyx_t_1, __pyx_t_23};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_22) {
          __Pyx_GIVEREF(__pyx_t_22); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_22); __pyx_t_22 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_13, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_23);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_13, __pyx_t_23);
        __pyx_t_1 = 0;
        __pyx_t_23 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_tocsr); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      if (__pyx_t_6) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 287, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_mat, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Orange/data/_io.pyx":288
 *         if len(t_data):
 *             mat = sp.coo_matrix((t_data, (t_rows, t_cols)), (line, len(ll))).tocsr()
 *             mat.sort_indices()             # <<<<<<<<<<<<<<
 *         else:
 *             mat = None
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_mat, __pyx_n_s_sort_indices); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      if (__pyx_t_6) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 288, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 288, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Orange/data/_io.pyx":286
 *              (Y_data, Y_rows, Y_cols, class_indices),
 *              (M_data, M_rows, M_cols, meta_indices)):
 *         if len(t_data):             # <<<<<<<<<<<<<<
 *             mat = sp.coo_matrix((t_data, (t_rows, t_cols)), (line, len(ll))).tocsr()
 *             mat.sort_indices()
 */
      goto __pyx_L66;
    }

    /* "Orange/data/_io.pyx":290
 *             mat.sort_indices()
 *         else:
 *             mat = None             # <<<<<<<<<<<<<<
 *         res.append(mat)
 * 
 */
    /*else*/ {
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_mat, Py_None);
    }
    __pyx_L66:;

    /* "Orange/data/_io.pyx":291
 *         else:
 *             mat = None
 *         res.append(mat)             # <<<<<<<<<<<<<<
 * 
 *     return tuple(res) + (attr_indices, class_indices, meta_indices)
 */
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_res, __pyx_v_mat); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 291, __pyx_L1_error)

    /* "Orange/data/_io.pyx":282
 * 
 *     res = []
 *     for t_data, t_rows, t_cols, ll in (             # <<<<<<<<<<<<<<
 *              (X_data, X_rows, X_cols, attr_indices),
 *              (Y_data, Y_rows, Y_cols, class_indices),
 */
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "Orange/data/_io.pyx":293
 *         res.append(mat)
 * 
 *     return tuple(res) + (attr_indices, class_indices, meta_indices)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyList_AsTuple(__pyx_v_res); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_attr_indices);
  __Pyx_GIVEREF(__pyx_v_attr_indices);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_attr_indices);
  __Pyx_INCREF(__pyx_v_class_indices);
  __Pyx_GIVEREF(__pyx_v_class_indices);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_class_indices);
  __Pyx_INCREF(__pyx_v_meta_indices);
  __Pyx_GIVEREF(__pyx_v_meta_indices);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_meta_indices);
  __pyx_t_10 = PyNumber_Add(__pyx_t_8, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "Orange/data/_io.pyx":16
 * 
 * 
 * def sparse_read_float(fname):             # <<<<<<<<<<<<<<
 *     cdef:
 *         State state
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_AddTraceback("Orange.data._io.sparse_read_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_X_data);
  __Pyx_XDECREF(__pyx_v_X_rows);
  __Pyx_XDECREF(__pyx_v_X_cols);
  __Pyx_XDECREF(__pyx_v_Y_data);
  __Pyx_XDECREF(__pyx_v_Y_rows);
  __Pyx_XDECREF(__pyx_v_Y_cols);
  __Pyx_XDECREF(__pyx_v_M_data);
  __Pyx_XDECREF(__pyx_v_M_rows);
  __Pyx_XDECREF(__pyx_v_M_cols);
  __Pyx_XDECREF(__pyx_v_attr_indices);
  __Pyx_XDECREF(__pyx_v_class_indices);
  __Pyx_XDECREF(__pyx_v_meta_indices);
  __Pyx_XDECREF(__pyx_v_t_names);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_t_data);
  __Pyx_XDECREF(__pyx_v_t_rows);
  __Pyx_XDECREF(__pyx_v_t_cols);
  __Pyx_XDECREF(__pyx_v_ll);
  __Pyx_XDECREF(__pyx_v_mat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__io(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__io},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_io",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 1},
  {&__pyx_kp_u_File_cannot_be_opened, __pyx_k_File_cannot_be_opened, sizeof(__pyx_k_File_cannot_be_opened), 0, 1, 0, 0},
  {&__pyx_n_s_IOError, __pyx_k_IOError, sizeof(__pyx_k_IOError), 0, 0, 1, 1},
  {&__pyx_n_s_M_cols, __pyx_k_M_cols, sizeof(__pyx_k_M_cols), 0, 0, 1, 1},
  {&__pyx_n_s_M_data, __pyx_k_M_data, sizeof(__pyx_k_M_data), 0, 0, 1, 1},
  {&__pyx_n_s_M_rows, __pyx_k_M_rows, sizeof(__pyx_k_M_rows), 0, 0, 1, 1},
  {&__pyx_n_s_Orange_data__io, __pyx_k_Orange_data__io, sizeof(__pyx_k_Orange_data__io), 0, 0, 1, 1},
  {&__pyx_kp_s_Orange_data__io_pyx, __pyx_k_Orange_data__io_pyx, sizeof(__pyx_k_Orange_data__io_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_X_cols, __pyx_k_X_cols, sizeof(__pyx_k_X_cols), 0, 0, 1, 1},
  {&__pyx_n_s_X_data, __pyx_k_X_data, sizeof(__pyx_k_X_data), 0, 0, 1, 1},
  {&__pyx_n_s_X_rows, __pyx_k_X_rows, sizeof(__pyx_k_X_rows), 0, 0, 1, 1},
  {&__pyx_n_s_Y_cols, __pyx_k_Y_cols, sizeof(__pyx_k_Y_cols), 0, 0, 1, 1},
  {&__pyx_n_s_Y_data, __pyx_k_Y_data, sizeof(__pyx_k_Y_data), 0, 0, 1, 1},
  {&__pyx_n_s_Y_rows, __pyx_k_Y_rows, sizeof(__pyx_k_Y_rows), 0, 0, 1, 1},
  {&__pyx_n_s_atom, __pyx_k_atom, sizeof(__pyx_k_atom), 0, 0, 1, 1},
  {&__pyx_n_s_atome, __pyx_k_atome, sizeof(__pyx_k_atome), 0, 0, 1, 1},
  {&__pyx_n_s_atomp, __pyx_k_atomp, sizeof(__pyx_k_atomp), 0, 0, 1, 1},
  {&__pyx_n_s_attr_index, __pyx_k_attr_index, sizeof(__pyx_k_attr_index), 0, 0, 1, 1},
  {&__pyx_n_s_attr_indices, __pyx_k_attr_indices, sizeof(__pyx_k_attr_indices), 0, 0, 1, 1},
  {&__pyx_n_s_b_atom, __pyx_k_b_atom, sizeof(__pyx_k_b_atom), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_ci, __pyx_k_ci, sizeof(__pyx_k_ci), 0, 0, 1, 1},
  {&__pyx_n_s_class_indices, __pyx_k_class_indices, sizeof(__pyx_k_class_indices), 0, 0, 1, 1},
  {&__pyx_kp_u_classes_should_follow_attribute, __pyx_k_classes_should_follow_attribute, sizeof(__pyx_k_classes_should_follow_attribute), 0, 1, 0, 0},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_col, __pyx_k_col, sizeof(__pyx_k_col), 0, 0, 1, 1},
  {&__pyx_n_s_col_kind, __pyx_k_col_kind, sizeof(__pyx_k_col_kind), 0, 0, 1, 1},
  {&__pyx_n_s_coo_matrix, __pyx_k_coo_matrix, sizeof(__pyx_k_coo_matrix), 0, 0, 1, 1},
  {&__pyx_n_s_cur_line, __pyx_k_cur_line, sizeof(__pyx_k_cur_line), 0, 0, 1, 1},
  {&__pyx_n_s_decs, __pyx_k_decs, sizeof(__pyx_k_decs), 0, 0, 1, 1},
  {&__pyx_kp_u_duplicated_semi_colons, __pyx_k_duplicated_semi_colons, sizeof(__pyx_k_duplicated_semi_colons), 0, 1, 0, 0},
  {&__pyx_kp_u_empty_value_name, __pyx_k_empty_value_name, sizeof(__pyx_k_empty_value_name), 0, 1, 0, 0},
  {&__pyx_kp_u_end_of_file_in_escape_sequence, __pyx_k_end_of_file_in_escape_sequence, sizeof(__pyx_k_end_of_file_in_escape_sequence), 0, 1, 0, 0},
  {&__pyx_kp_u_end_of_file_within_a_quoted_val, __pyx_k_end_of_file_within_a_quoted_val, sizeof(__pyx_k_end_of_file_within_a_quoted_val), 0, 1, 0, 0},
  {&__pyx_kp_u_end_of_line_in_escape_sequence, __pyx_k_end_of_line_in_escape_sequence, sizeof(__pyx_k_end_of_line_in_escape_sequence), 0, 1, 0, 0},
  {&__pyx_kp_u_end_of_line_within_a_quoted_val, __pyx_k_end_of_line_within_a_quoted_val, sizeof(__pyx_k_end_of_line_within_a_quoted_val), 0, 1, 0, 0},
  {&__pyx_n_s_endc, __pyx_k_endc, sizeof(__pyx_k_endc), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_f_eof, __pyx_k_f_eof, sizeof(__pyx_k_f_eof), 0, 0, 1, 1},
  {&__pyx_n_s_fname, __pyx_k_fname, sizeof(__pyx_k_fname), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_in_line, __pyx_k_in_line, sizeof(__pyx_k_in_line), 0, 0, 1, 1},
  {&__pyx_kp_u_invalid_value, __pyx_k_invalid_value, sizeof(__pyx_k_invalid_value), 0, 1, 0, 0},
  {&__pyx_n_s_line, __pyx_k_line, sizeof(__pyx_k_line), 0, 0, 1, 1},
  {&__pyx_n_s_ll, __pyx_k_ll, sizeof(__pyx_k_ll), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mat, __pyx_k_mat, sizeof(__pyx_k_mat), 0, 0, 1, 1},
  {&__pyx_n_s_meta_indices, __pyx_k_meta_indices, sizeof(__pyx_k_meta_indices), 0, 0, 1, 1},
  {&__pyx_kp_u_missing_value_name, __pyx_k_missing_value_name, sizeof(__pyx_k_missing_value_name), 0, 1, 0, 0},
  {&__pyx_n_s_not_in_atom, __pyx_k_not_in_atom, sizeof(__pyx_k_not_in_atom), 0, 0, 1, 1},
  {&__pyx_kp_u_quoted_value_should_be_followed, __pyx_k_quoted_value_should_be_followed, sizeof(__pyx_k_quoted_value_should_be_followed), 0, 1, 0, 0},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_sparse, __pyx_k_scipy_sparse, sizeof(__pyx_k_scipy_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_sort_indices, __pyx_k_sort_indices, sizeof(__pyx_k_sort_indices), 0, 0, 1, 1},
  {&__pyx_n_s_sp, __pyx_k_sp, sizeof(__pyx_k_sp), 0, 0, 1, 1},
  {&__pyx_n_s_sparse_read_float, __pyx_k_sparse_read_float, sizeof(__pyx_k_sparse_read_float), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_t_cols, __pyx_k_t_cols, sizeof(__pyx_k_t_cols), 0, 0, 1, 1},
  {&__pyx_n_s_t_data, __pyx_k_t_data, sizeof(__pyx_k_t_data), 0, 0, 1, 1},
  {&__pyx_n_s_t_names, __pyx_k_t_names, sizeof(__pyx_k_t_names), 0, 0, 1, 1},
  {&__pyx_n_s_t_rows, __pyx_k_t_rows, sizeof(__pyx_k_t_rows), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tocsr, __pyx_k_tocsr, sizeof(__pyx_k_tocsr), 0, 0, 1, 1},
  {&__pyx_kp_u_unrecognized_escape_sequence, __pyx_k_unrecognized_escape_sequence, sizeof(__pyx_k_unrecognized_escape_sequence), 0, 1, 0, 0},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_kp_u_value_name_too_long, __pyx_k_value_name_too_long, sizeof(__pyx_k_value_name_too_long), 0, 1, 0, 0},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_IOError = __Pyx_GetBuiltinName(__pyx_n_s_IOError); if (!__pyx_builtin_IOError) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 87, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "Orange/data/_io.pyx":16
 * 
 * 
 * def sparse_read_float(fname):             # <<<<<<<<<<<<<<
 *     cdef:
 *         State state
 */
  __pyx_tuple__2 = PyTuple_Pack(39, __pyx_n_s_fname, __pyx_n_s_state, __pyx_n_s_c, __pyx_n_s_ci, __pyx_n_s_not_in_atom, __pyx_n_s_col, __pyx_n_s_line, __pyx_n_s_cur_line, __pyx_n_s_in_line, __pyx_n_s_atom, __pyx_n_s_atomp, __pyx_n_s_atome, __pyx_n_s_endc, __pyx_n_s_f_eof, __pyx_n_s_attr_index, __pyx_n_s_value, __pyx_n_s_decs, __pyx_n_s_col_kind, __pyx_n_s_X_data, __pyx_n_s_X_rows, __pyx_n_s_X_cols, __pyx_n_s_Y_data, __pyx_n_s_Y_rows, __pyx_n_s_Y_cols, __pyx_n_s_M_data, __pyx_n_s_M_rows, __pyx_n_s_M_cols, __pyx_n_s_attr_indices, __pyx_n_s_class_indices, __pyx_n_s_meta_indices, __pyx_n_s_f, __pyx_n_s_t_names, __pyx_n_s_b_atom, __pyx_n_s_res, __pyx_n_s_t_data, __pyx_n_s_t_rows, __pyx_n_s_t_cols, __pyx_n_s_ll, __pyx_n_s_mat); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(1, 0, 39, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Orange_data__io_pyx, __pyx_n_s_sparse_read_float, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_setdefault.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((optimize("Os")))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_io(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_io(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__io(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__io(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec__io(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__io(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_io", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_Orange__data___io) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "Orange.data._io")) {
      if (unlikely(PyDict_SetItemString(modules, "Orange.data._io", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "Orange/data/_io.pyx":1
 * import scipy.sparse as sp             # <<<<<<<<<<<<<<
 * 
 * from libc.stdio cimport fopen, fclose, fgetc, ungetc, EOF, FILE
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_);
  __Pyx_GIVEREF(__pyx_n_s_);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_sparse, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sp, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Orange/data/_io.pyx":16
 * 
 * 
 * def sparse_read_float(fname):             # <<<<<<<<<<<<<<
 *     cdef:
 *         State state
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6Orange_4data_3_io_1sparse_read_float, NULL, __pyx_n_s_Orange_data__io); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sparse_read_float, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Orange/data/_io.pyx":1
 * import scipy.sparse as sp             # <<<<<<<<<<<<<<
 * 
 * from libc.stdio cimport fopen, fclose, fgetc, ungetc, EOF, FILE
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init Orange.data._io", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init Orange.data._io");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST);
    }
#endif
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       CYTHON_UNUSED int is_safe_type) {
    PyObject* value;
#if PY_VERSION_HEX >= 0x030400A0
    if ((1)) {
        value = PyDict_SetDefault(d, key, default_value);
        if (unlikely(!value)) return NULL;
        Py_INCREF(value);
#else
    if (is_safe_type == 1 || (is_safe_type == -1 &&
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
            (PyUnicode_CheckExact(key) || PyString_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItemWithError(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#else
            (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#endif
#endif
    } else {
        value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_setdefault, d, key, default_value);
    }
    return value;
}

/* None */
    static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* GetException */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if PY_VERSION_HEX >= 0x030700A2
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = local_type;
    tstate->exc_state.exc_value = local_value;
    tstate->exc_state.exc_traceback = local_tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A2
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = *type;
    tstate->exc_state.exc_value = *value;
    tstate->exc_state.exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* SaveResetException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if PY_VERSION_HEX >= 0x030700A2
    *type = tstate->exc_state.exc_type;
    *value = tstate->exc_state.exc_value;
    *tb = tstate->exc_state.exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A2
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = type;
    tstate->exc_state.exc_value = value;
    tstate->exc_state.exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* RaiseTooManyValuesToUnpack */
      static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
      static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
      static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* GetModuleGlobalName */
      static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    if (likely(result)) {
        Py_INCREF(result);
    } else if (unlikely(PyErr_Occurred())) {
        result = NULL;
    } else {
#else
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCallNoArg */
          #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* Import */
            static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* CLineInTraceback */
            #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
            static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
            #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value) {
    const char neg_one = (char) -1, const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
            #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
            static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
            #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
            static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
            static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
